import json
import logging
import pickle
from collections.abc import Callable, Generator, Iterable, Iterator
from contextlib import suppress
from itertools import combinations, pairwise
from math import comb
from pathlib import Path
from pprint import pformat
from typing import Any, Literal

import h5py
import numpy as np

from . import Blob, Fragment, GlobalFragment
from .utils import (
    clean_attrs,
    conf,
    deprecated,
    load_id_images,
    open_track,
    resolve_path,
    track,
)


class ListOfFragments:
    """Contains all the instances of the class :class:`Fragment`."""

    accumulable_individual_fragments: set[int]

    not_accumulable_individual_fragments: set[int]

    id_to_exclusive_roi: np.ndarray
    "Maps identities (from 0 to n_animals-1) to their exclusive ROI (-1 meaning no ROI)"

    n_animals: int

    fragments: list[Fragment]

    id_images_file_paths: list[Path]

    def __init__(
        self,
        fragments: list[Fragment],
        id_images_file_paths: list[Path],
        number_of_animals: int,
    ) -> None:
        """Initialize the ListOfFragments manually from a list of :class:`Fragment`.
        This method should only be called from :meth:`from_fragmented_blobs`. You can
        load a ListOfFragments from a file by using the method :meth:`load`.

        Parameters
        ----------
        fragments : list[Fragment]
            List of instances of the class :class:`Fragment`.
        id_images_file_paths : list[Path]
            List of paths to the files where the identification images are stored.
        number_of_animals : int
            Number of animals in the video
        """

        # Assert fragments are sorted
        assert all(i == frag.identifier for i, frag in enumerate(fragments))
        self.n_animals = number_of_animals
        self.fragments = fragments
        self.id_images_file_paths = id_images_file_paths
        self.connect_coexisting_fragments()
        self.id_to_exclusive_roi = np.full(self.n_animals, -1)

    def __iter__(self) -> Iterator[Fragment]:
        return iter(self.fragments)

    def __len__(self) -> int:
        return len(self.fragments)

    @classmethod
    def load(
        cls, file_path: Path | str, reconnect: bool = True, verbose: bool = True
    ) -> "ListOfFragments":
        """Loads a previously saved list of fragments (see :meth:`save`) from a JSON file

        Parameters
        ----------
        file_path : Path | str
            Path to the JSON file containing
        reconnect : bool, optional
            If True, the :attr:`Fragments.coexisting_individual_fragments` attribute,
            which is not saved in the JSON file, is computed and populated, by default True

        Returns
        -------
        ListOfFragments
            Loaded instance
        """
        file_path = resolve_path(file_path)
        logging.info(f"Loading ListOfFragments from {file_path}", stacklevel=2)

        if not file_path.is_file() and file_path.with_suffix(".pickle").is_file():
            # <=5.1.3 compatibility
            with file_path.with_suffix(".pickle").open("rb") as f:
                pickle.load(f).save(file_path)

        list_of_fragments = cls.__new__(cls)

        with open_track(file_path.with_suffix(".json"), verbose=verbose) as file:
            json_data: dict = json.load(file)

        list_of_fragments.accumulable_individual_fragments = set(
            json_data.get("accumulable_individual_fragments", [])
        )
        list_of_fragments.not_accumulable_individual_fragments = set(
            json_data.get("not_accumulable_individual_fragments", [])
        )
        if "number_of_animals" in json_data:
            list_of_fragments.n_animals = json_data["number_of_animals"]
        if "n_animals" in json_data:
            list_of_fragments.n_animals = json_data["n_animals"]

        list_of_fragments.id_images_file_paths = list(
            map(Path, json_data["id_images_file_paths"])
        )

        list_of_fragments.fragments = [
            Fragment.from_json(frag_data) for frag_data in json_data["fragments"]
        ]

        for fragment in list_of_fragments:
            if (
                fragment.identifier
                in list_of_fragments.accumulable_individual_fragments
            ):
                fragment.accumulable = True
            elif (
                fragment.identifier
                in list_of_fragments.not_accumulable_individual_fragments
            ):
                fragment.accumulable = False

        if reconnect:
            list_of_fragments.connect_coexisting_fragments(verbose=verbose)

        with suppress(AttributeError):
            list_of_fragments.id_to_exclusive_roi = np.asarray(
                json_data.get(
                    "id_to_exclusive_roi", np.full(list_of_fragments.n_animals, -1)
                )
            )

        new_location = file_path.parent.parent / "identification_images"
        if new_location != list_of_fragments.id_images_file_paths[0].parent and all(
            (new_location / file.name).is_file()
            for file in list_of_fragments.id_images_file_paths
        ):
            logging.info("Updating List Of Fragments id_images_file_paths")
            list_of_fragments.id_images_file_paths = [
                new_location / file.name
                for file in list_of_fragments.id_images_file_paths
            ]

        return list_of_fragments

    @classmethod
    def from_fragmented_blobs(
        cls,
        all_blobs: Iterable[Blob],
        number_of_animals: int,
        id_images_file_paths: list[Path],
    ) -> "ListOfFragments":
        """Generate a list of instances of :class:`Fragment` collecting
        all the fragments in the video.

        Parameters
        ----------
        all_blobs : Iterable[Blob]
            An iterable of the blob objects which have to be sorted by
            :attr:`Blob.frame_number`, for example :attr:`ListOfBlobs.all_blobs`.
        number_of_animals : int
            Number of animals to track as defined by the user
        id_images_file_paths : list[Path]
            List of paths to the files where the identification images are stored.

        Returns
        -------
        ListOfFragments
            The List of Fragments computed from the blobs.
        """

        fragments: list[Fragment] = []
        used_fragment_identifiers: set[int] = set()

        logging.info("Creating list of fragments")
        for blob in all_blobs:
            current_fragment_identifier = blob.fragment_identifier
            if current_fragment_identifier in used_fragment_identifiers:
                continue
            images = [blob.id_image_index]
            centroids = [blob.centroid]
            episodes = [blob.episode]
            start = blob.frame_number
            exclusive_roi = blob.exclusive_roi
            current = blob

            while (
                current.n_next > 0
                and current.next[0].fragment_identifier == current_fragment_identifier
            ):
                current = current.next[0]
                images.append(current.id_image_index)
                centroids.append(current.centroid)
                episodes.append(current.episode)

            end = current.frame_number

            fragment = Fragment(
                current_fragment_identifier,
                start,
                end + 1,  # it is not inclusive
                images,
                centroids,
                episodes,
                blob.is_an_individual,
                exclusive_roi,
            )
            used_fragment_identifiers.add(current_fragment_identifier)
            fragments.append(fragment)
        return cls(fragments, id_images_file_paths, number_of_animals)

    def save(self, file_path: Path | str) -> None:
        """Save the list of fragments into a JSON file

        Parameters
        ----------
        file_path : Path | str
            Path where the instance of the object will be stored.
        """
        file_path = resolve_path(file_path)
        if file_path.is_dir():
            file_path /= "list_of_fragments.json"
        logging.info(f"Saving ListOfFragments as {file_path}", stacklevel=2)
        file_path.parent.mkdir(exist_ok=True)

        with open_track(file_path, "w") as file:
            json.dump(self, file, cls=FragmentsEncoder, indent=4)

    @property
    def number_of_fragments(self) -> int:
        return len(self.fragments)

    @property
    def individual_fragments(self) -> Generator[Fragment, None, None]:
        return (frag for frag in self if frag.is_an_individual)

    def get_connectivity(
        self,
        min_fragment_length: int = conf.MIN_N_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION,
    ) -> float:
        """Computes the connectivity of the fragments."""
        long_fragments = [
            frag
            for frag in self.individual_fragments
            if frag.n_images >= min_fragment_length
        ]

        coexistence_count = 0
        for fragment in long_fragments:
            for coex_frag in fragment.coexisting_individual_fragments:
                if (
                    coex_frag.is_an_individual
                    and coex_frag.n_images >= min_fragment_length
                ):
                    coexistence_count += 1

        return coexistence_count / (len(long_fragments) * (self.n_animals - 1))

    # TODO: if the resume feature is not active, this does not make sense|
    def reset(self, roll_back_to: Literal["fragmentation", "accumulation"]) -> None:
        """Resets all the fragment to a given processing step.

        Parameters
        ----------
        roll_back_to : str
            Name of the step at which the fragments should be reset.
            It can be 'fragmentation', 'accumulation' or

        See Also
        --------
        :meth:`Fragment.reset`
        """
        logging.info(f"Resetting ListOfFragments to '{roll_back_to}'", stacklevel=2)
        for fragment in self:
            fragment.reset(roll_back_to, self.n_animals)

    @property
    def n_images_in_global_fragments(self) -> int:
        """Number of images available in global fragments
        (without repetitions)"""
        return sum(
            frag.n_images
            for frag in self
            if frag.identifier in self.accumulable_individual_fragments
        )

    @property
    def ratio_of_images_used_for_training(self) -> float:
        """Returns the ratio of images used for training over the number of
        available images"""
        return (
            sum(fragment.n_images for fragment in self if fragment.used_for_training)
            / self.n_images_in_global_fragments
        )

    def build_exclusive_rois(self) -> dict[str, set[int]]:
        """Builds `id_to_exclusive_roi` and returns a more readable version
        intended to be saved in Session.identities_groups"""

        # build id_to_exclusive_roi
        for fragment in self:
            if fragment.temporary_id is not None:
                self.id_to_exclusive_roi[fragment.temporary_id] = fragment.exclusive_roi

        # build identity groups to save in Session
        id_groups: dict[str, set] = {}
        for id, roi in enumerate(self.id_to_exclusive_roi):
            if roi == -1:
                continue
            roi_name = f"Region_{roi}"
            if roi_name in id_groups:
                id_groups[roi_name].add(id + 1)
            else:
                id_groups[roi_name] = {id + 1}
        if id_groups:
            logging.info("Identity groups by exclusive ROIs:\n%s", pformat(id_groups))

        return id_groups

    def compute_P2_vectors(self) -> None:
        """Computes the P2_vector associated to every individual fragment. See
        :meth:`Fragment.compute_P2_vector`
        """
        for fragment in self.individual_fragments:
            fragment.compute_P2_vector(self.n_animals)

    def get_fragments_to_identify(self) -> Iterator[Fragment]:
        """Yields all the :class:`Fragment` instances left non identified after
        the cascade of training and identification protocols sorted by the
        certainty computed with P2. See :attr:Fragment.certainty_P2`

        Yields
        ------
        Iterator[Fragment]
            Non identified Fragments sorted by P2 certainty
        """
        frags_to_identity = [
            frag for frag in self.individual_fragments if frag.identity is None
        ]
        while frags_to_identity:
            yield max(frags_to_identity, key=lambda frag: frag.certainty_P2)
            frags_to_identity = [
                frag for frag in self.individual_fragments if frag.identity is None
            ]

    def update_id_images_dataset(self) -> None:
        """Updates the identification images files with the identity assigned
        to each fragment during the tracking process.
        """
        logging.info("Updating identities in identification images files")

        identities = []
        for path in self.id_images_file_paths:
            with h5py.File(path, "r") as file:
                identities.append(np.full(len(file["id_images"]), 0))  # type: ignore

        for fragment in self:
            if fragment.used_for_training:
                for image, episode in fragment.image_locations:
                    identities[episode][image] = fragment.identity

        for path, identities_in_episode in zip(self.id_images_file_paths, identities):
            try:
                with h5py.File(path, "r+") as file:
                    dataset = file.require_dataset(
                        "identities", shape=len(identities_in_episode), dtype=int
                    )
                    dataset[:] = identities_in_episode
            except BlockingIOError as exc:
                # Some MacOS crash with
                # BlockingIOError: [Errno 35] Unable to open file (unable to lock file, errno = 35, error message = 'Resource temporarily unavailable')
                logging.error(f"Failed to update identities in {path} due to: {exc}")

    def get_ordered_list_of_fragments(
        self, scope: Literal["to_the_past", "to_the_future"], specific_frame: int
    ) -> list[Fragment]:
        """Sorts the fragments starting from the frame number
        `first_frame_first_global_fragment`. According to `scope` the sorting
        is done either "to the future" of "to the past" with respect to
        `first_frame_first_global_fragment`

        Parameters
        ----------
        scope : str
            either "to_the_past" or "to_the_future"
        first_frame_first_global_fragment : int
            frame number corresponding to the first frame in which all the
            individual fragments coexist in the first global fragment using
            in an iteration of the fingerprint protocol cascade

        Returns
        -------
        list
            list of sorted fragments

        """
        if scope == "to_the_past":
            fragments_to_the_past = filter(
                lambda frag: frag.end_frame <= specific_frame, self.fragments
            )
            return sorted(
                fragments_to_the_past, key=lambda x: x.end_frame, reverse=True
            )
        if scope == "to_the_future":
            fragments_to_the_future = filter(
                lambda frag: frag.start_frame >= specific_frame, self.fragments
            )
            return sorted(fragments_to_the_future, key=lambda x: x.start_frame)
        raise ValueError(scope)

    def load_images_in_memory(
        self, condition: Callable[[Fragment], bool] | None = None
    ):
        """Loads Fragment's images in memory from id_images_file_paths.
        It only takes into account the fragments satisfying the given condition.
        Used outside idtracker.ai"""
        image_locations = []
        fragments = list(filter(condition, self)) if condition is not None else self
        for frag in fragments:
            image_locations += frag.image_locations
        all_images = load_id_images(self.id_images_file_paths, image_locations)
        counter = 0
        for frag in fragments:
            frag.loaded_images = all_images[counter : counter + frag.n_images]
            counter += frag.n_images
        assert counter == len(all_images)

    def connect_coexisting_fragments(self, verbose=True) -> None:
        "Populates :attr:`Fragment.coexisting_individual_fragments`"
        for fragment in self:
            fragment.coexisting_individual_fragments = []

        # If fragments are sorted by starting frame, we have a nice optimization
        if all(
            frag_A.start_frame <= frag_B.start_frame and frag_A.identifier == i
            for i, (frag_A, frag_B) in enumerate(pairwise(self.fragments))
        ):
            for fragment_A in track(
                self.fragments, "Connecting coexisting fragments", verbose=verbose
            ):
                for fragment_B in self.fragments[fragment_A.identifier + 1 :]:
                    if fragment_A.coexist_with(fragment_B):
                        if fragment_A.is_an_individual:
                            fragment_B.coexisting_individual_fragments.append(fragment_A)  # type: ignore
                        if fragment_B.is_an_individual:
                            fragment_A.coexisting_individual_fragments.append(fragment_B)  # type: ignore
                    if fragment_B.start_frame > fragment_A.end_frame:
                        break
            return

        # brute force
        logging.warning("List of Fragments is not sorted")
        for fragment_A, fragment_B in track(
            combinations(self.fragments, 2),
            "Connecting coexisting fragments",
            comb(len(self.fragments), 2),
            verbose=verbose,
        ):
            if fragment_A.coexist_with(fragment_B):
                if fragment_A.is_an_individual:
                    fragment_B.coexisting_individual_fragments.append(fragment_A)  # type: ignore
                if fragment_B.is_an_individual:
                    fragment_A.coexisting_individual_fragments.append(fragment_B)  # type: ignore

    def manage_accumulable_non_accumulable_fragments(
        self,
        accumulable_global_fragments: list[GlobalFragment],
        non_accumulable_global_fragments: list[GlobalFragment],
    ):
        """Gets the unique identifiers associated to individual fragments that
        can be accumulated.

        Parameters
        ----------
        list_of_global_fragments : :class:`ListOfGlobalFragments`
            Object collecting the global fragment objects (instances of the
            class :class:`GlobalFragment`) detected in the
            entire video.

        """
        self.accumulable_individual_fragments = {
            identifier
            for glob_frag in accumulable_global_fragments
            for identifier in glob_frag.fragments_identifiers
        }
        self.not_accumulable_individual_fragments = {
            identifier
            for glob_frag in non_accumulable_global_fragments
            for identifier in glob_frag.fragments_identifiers
        } - self.accumulable_individual_fragments

        for fragment in self:
            if fragment.identifier in self.accumulable_individual_fragments:
                fragment.accumulable = True
            elif fragment.identifier in self.not_accumulable_individual_fragments:
                fragment.accumulable = False

    @property
    def number_of_crossing_fragments(self) -> int:
        return sum(fragment.is_a_crossing for fragment in self)

    @property
    def number_of_individual_fragments(self) -> int:
        return sum(1 for _ in self.individual_fragments)

    def get_stats(self) -> dict[str, int]:
        """Collects the following counters from the fragments.

        * fragments
        * crossing_fragments
        * individual_fragments
        * individual_fragments_not_in_a_global_fragment
        * accumulable_individual_fragments
        * not_accumulable_individual_fragments
        * globally_accumulated_individual_fragments
        * partially_accumulated_individual_fragments
        * blobs
        * crossing_blobs
        * individual_blobs
        * individual_blobs_not_in_a_global_fragment
        * accumulable_individual_blobs
        * not_accumulable_individual_blobs
        * globally_accumulated_individual_blobs
        * partially_accumulated_individual_blobs

        Returns
        -------
        dict
            Dictionary with the counters mentioned above

        """

        stats: dict[str, int] = {
            "fragments": self.number_of_fragments,
            "crossing_fragments": self.number_of_crossing_fragments,
            "individual_fragments": self.number_of_individual_fragments,
            "individual_fragments_not_in_a_global_fragment": sum(
                not frag.is_in_a_global_fragment for frag in self.individual_fragments
            ),
            "accumulable_individual_fragments": len(
                self.accumulable_individual_fragments
            ),
            "not_accumulable_individual_fragments": len(
                self.not_accumulable_individual_fragments
            ),
            "globally_accumulated_individual_fragments": sum(
                frag.accumulated_globally for frag in self.individual_fragments
            ),
            "partially_accumulated_individual_fragments": sum(
                frag.accumulated_partially for frag in self.individual_fragments
            ),
            "blobs": sum(frag.n_images for frag in self),
            "crossing_blobs": sum(frag.is_a_crossing * frag.n_images for frag in self),
            "individual_blobs": sum(
                frag.n_images for frag in self.individual_fragments
            ),
            "individual_blobs_not_in_a_global_fragment": sum(
                not frag.is_in_a_global_fragment * frag.n_images
                for frag in self.individual_fragments
            ),
            "accumulable_individual_blobs": sum(
                bool(frag.accumulable) * frag.n_images for frag in self
            ),
            "not_accumulable_individual_blobs": sum(
                (not frag.accumulable) * frag.n_images
                for frag in self
                if frag.accumulable is not None
            ),
            "globally_accumulated_individual_blobs": sum(
                frag.accumulated_globally * frag.n_images
                for frag in self.individual_fragments
            ),
            "partially_accumulated_individual_blobs": sum(
                frag.accumulated_partially * frag.n_images
                for frag in self.individual_fragments
            ),
        }

        log = "Final statistics:"
        for key, value in stats.items():
            log += f"\n  {value} {key.replace('_', ' ')}"
        logging.info(log)

        return stats

    def update_blobs(self, all_blobs: Iterable[Blob]) -> None:
        """Updates the blobs objects generated from the video with the
        attributes computed for each fragment

        Parameters
        ----------
        all_blobs : Iterable[Blob]
            All blobs to be updated
        """
        logging.info("Updating list of blobs from list of fragments")
        for blob in all_blobs:
            fragment = self.fragments[blob.fragment_identifier]
            blob.identity = fragment.identity
            blob.identity_corrected_solving_jumps = (
                fragment.identity_corrected_solving_jumps
            )
            blob.is_an_individual = fragment.is_an_individual
            if fragment.forced_crossing:
                blob.forced_crossing = True

    # Deprecated methods

    @deprecated(
        version="6.0.0",
        reason="Use :meth:`get_fragments_to_identify` instead",
        action="error",
    )
    def get_next_fragment_to_identify(self): ...

    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def get_number_of_unidentified_individual_fragments(): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def ratio_of_images_used_for_pretraining(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_individual_fragments_not_in_a_glob_fragment(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_accumulable_individual_fragments(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_not_accumulable_individual_fragments(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_blobs(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_crossing_blobs(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_individual_blobs(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_individual_blobs_not_in_a_global_fragment(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def fragments_not_accumulated(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_globally_accumulated_individual_fragments(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_partially_accumulated_individual_fragments(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_accumulable_individual_blobs(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_not_accumulable_individual_blobs(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_globally_accumulated_individual_blobs(self): ...
    @property
    @deprecated(
        version="6.0.0",
        reason="See :meth:`get_stats` to get the same information",
        action="error",
    )
    def number_of_partially_accumulated_individual_blobs(self): ...


class FragmentsEncoder(json.JSONEncoder):
    def default(self, o):
        match o:
            case Path():
                return str(o)

            case ListOfFragments():
                serial = o.__dict__.copy()
                serial["id_to_exclusive_roi"] = (
                    f"NotString{serial.get('id_to_exclusive_roi', np.array(())).tolist()}"
                )
                serial["accumulable_individual_fragments"] = (
                    f"NotString{list(serial.get('accumulable_individual_fragments', {}))}"
                )
                serial["not_accumulable_individual_fragments"] = (
                    f"NotString{list(serial.get('not_accumulable_individual_fragments', {}))}"
                )
                return serial

            case Fragment():
                clean_attrs(o)
                serial = o.__getstate__()

                serial["episodes"] = f"NotString{compress(o.episodes).tolist()}"

                for key in (
                    "frame_by_frame_velocity",
                    "start_position",
                    "end_position",
                ):
                    if key in serial:
                        serial[key] = f"NotString{np.round(serial[key], 2).tolist()}"

                for key in ("images", "P1_vector", "P2_vector", "ambiguous_identities"):
                    if key in serial:
                        serial[key] = f"NotString{serial[key].tolist()}"

                return serial
            case np.integer():
                return int(o)
            case np.bool_():
                return bool(o)
            case np.floating():
                return float(o)
            case _:
                return super().default(o)

    def iterencode(self, o: Any, _one_shot: bool = False) -> Iterator[str]:
        for encoded in super().iterencode(o, _one_shot):
            if encoded.startswith('"NotString'):
                # remove first and final '"NoIndent..."' and remove indents,
                yield encoded[10:-1]
            else:
                yield encoded


def compress(arr: np.ndarray) -> np.ndarray[Any, np.dtype[np.int_]]:
    """Compresses an integer 1D array by finding its repetitions.

    [0,0,0,1,1,2,2,2] -> [[0,1,2], [3,2,3]]"""
    if len(arr) == 0:
        return np.array([[], []], dtype=int)
    changing_indices: np.ndarray = np.diff(arr, prepend=-1, append=-1) != 0
    values = arr[changing_indices[:-1]]
    repetitions = np.diff((~changing_indices).cumsum()[changing_indices]) + 1
    return np.asarray((values, repetitions))



from collections.abc import Sequence
from typing import Literal

import numpy as np

from idtrackerai import Fragment, ListOfFragments, Session
from idtrackerai.utils import track


def none_max(*args):
    return max(filter(lambda item: item is not None, args))


def get_candidate_identities_by_minimum_speed(
    fragment: Fragment,
    fragments: Sequence[Fragment],
    available_identities: list[int],
    impossible_velocity_threshold: float,
) -> tuple[np.ndarray, np.ndarray]:
    """Computes the candidate identities for a given `fragment` taking into
    consideration the velocities needed to join the `fragment` with its neighbour
    fragments in the past and in the future

    Parameters
    ----------
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal
    fragments : list
        List with all the `Fragment` objects of the video
    available_identities : set
        Set with the available idenities for the `fragment`
    impossible_velocity_threshold : float
        If the velocity needed to link two fragments is higher than this threshold
        the identiy of one of the fragments is considerd to be wrong as it would be
        physically impossible for an animal to move so much. See `video.velocity_threshold`
        for each definition

    Returns
    -------
    candidate_identities_by_speed : nd.array
        Array with the identities that fullfill the `impossible_velocity_threshold`
        ordered from minimum to maximum velocity

    speed_of_candidate_identities : nd.array
        Array with the maximum velocity needed to link the given `fragment`
        with its neighbours assuming a given identity. Ordered from minimum to maximum
        velocity
    """
    speed_of_candidate_identities: list[float] = []
    current_identity = fragment.identity_corrected_solving_jumps
    for identity in available_identities:
        fragment.identity_corrected_solving_jumps = identity
        neighbour_fragment_past = fragment.get_neighbour_fragment(
            fragments, "to_the_past"
        )
        neighbour_fragment_future = fragment.get_neighbour_fragment(
            fragments, "to_the_future"
        )
        velocity_past = fragment.compute_border_velocity(neighbour_fragment_past)
        velocity_future = fragment.compute_border_velocity(neighbour_fragment_future)

        if velocity_past is None and velocity_future is None:
            speed_of_candidate_identities.append(impossible_velocity_threshold)
        else:
            speed_of_candidate_identities.append(
                none_max(velocity_past, velocity_future)
            )
    fragment.identity_corrected_solving_jumps = current_identity
    argsort_identities_by_speed = np.argsort(speed_of_candidate_identities)
    return (
        np.asarray(available_identities)[argsort_identities_by_speed],
        np.asarray(speed_of_candidate_identities)[argsort_identities_by_speed],
    )


def get_candidate_identities_above_random_P2(
    fragment: Fragment,
    fragments: Sequence[Fragment],
    non_available_identities: np.ndarray,
    available_identities: list[int],
    impossible_velocity_threshold: float,
    number_of_animals: int,
):
    """Computes the candidate identities of a `fragment` taking into
    consideration the probability of identification given by its
    `fragment.P2_vector`. An identity is a potential candidate if the
    probability of identification is above random.

    Parameters
    ----------
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of
        overlapping blobs that are considered to be the same animal
    fragments : list
        List with all the `Fragment` objects of the video
    non_available_identities : nd.array
        Array with the non available identities for the `fragment`
    available_identities : set
        Set with the available idenities for the `fragment`
    impossible_velocity_threshold : float
        If the velocity needed to link two fragments is higher than this
        threshold the identiy of one of the fragments is considerd to be
        wrong as it would be physically impossible for an animal to move so
        much. See `video.velocity_threshold` for each definition

    Returns
    -------
    candidate_identities_by_speed : nd.array
        Array with the identities that fullfill the
        `impossible_velocity_threshold` ordered from minimum to maximum
        velocity

    speed_of_candidate_identities : nd.array
        Array with the maximum velocity needed to link the given `fragment`
        with its neighbours assuming a given identity. Ordered from
        minimum to maximum velocity

    See Also
    --------
    Fragment
    get_candidate_identities_by_minimum_speed

    """
    P2_vector = fragment.P2_vector
    assert P2_vector is not None
    if len(non_available_identities) > 0:
        P2_vector[non_available_identities - 1] = 0
    if all(P2_vector == 0):
        candidate_identities_speed, _ = get_candidate_identities_by_minimum_speed(
            fragment, fragments, available_identities, impossible_velocity_threshold
        )
        return candidate_identities_speed

    if fragment.n_images == 1:
        random_threshold = 1 / number_of_animals
    else:
        random_threshold = 1 / fragment.n_images
    return (P2_vector > random_threshold).nonzero()[0] + 1


def reassign(
    fragment: Fragment,
    list_of_fragments: ListOfFragments,
    impossible_velocity_threshold: float,
):
    """Reassigns the identity of a given `fragment` considering the identity of the
    `fragments` coexisting with it and the `impossible_velocity_threshold`

    Parameters
    ----------
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of overlapping
        blobs that are considered to be the same animal
    fragments : list
        List with all the `Fragment` objects of the video
    impossible_velocity_threshold : float
        If the velocity needed to link two fragments is higher than this threshold
        the identiy of one of the fragments is considerd to be wrong as it would be
        physically impossible for an animal to move so much. See `video.velocity_threshold`
        for each definition

    See Also
    --------
    :class:`Fragment`
    :meth:`get_available_and_non_available_identities`
    :meth:`get_candidate_identities_by_minimum_speed`
    :meth:`get_candidate_identities_above_random_P2`

    """

    coexisting_identities = {
        coexisting_fragment.assigned_identities[0]
        for coexisting_fragment in fragment.coexisting_individual_fragments
    } - {0, None}

    identities_outside_exclusive_roi: set[int] = {
        int(x)
        for x in (
            np.argwhere(list_of_fragments.id_to_exclusive_roi != fragment.exclusive_roi)
            + 1
        ).ravel()
    }
    all_identities = set(range(1, list_of_fragments.n_animals + 1))

    non_available_identities = coexisting_identities | identities_outside_exclusive_roi
    available_identities = all_identities - non_available_identities

    assert fragment.assigned_identities[0] not in non_available_identities

    non_available_identities = np.asarray(list(non_available_identities))

    if len(available_identities) == 1:
        candidate_id = available_identities.pop()
    else:
        available_identities = list(available_identities)
        candidate_identities_speed, speed_of_candidate_identities = (
            get_candidate_identities_by_minimum_speed(
                fragment,
                list_of_fragments.fragments,
                available_identities,
                impossible_velocity_threshold,
            )
        )
        candidate_identities_P2 = get_candidate_identities_above_random_P2(
            fragment,
            list_of_fragments.fragments,
            non_available_identities,
            available_identities,
            impossible_velocity_threshold,
            list_of_fragments.n_animals,
        )
        candidate_identities: list[int] = []
        candidate_speeds: list[float] = []
        for candidate_id, candidate_speed in zip(
            candidate_identities_speed, speed_of_candidate_identities
        ):
            if candidate_id in candidate_identities_P2:
                candidate_identities.append(candidate_id)
                candidate_speeds.append(candidate_speed)
        if not candidate_identities:
            candidate_id = 0
        elif len(candidate_identities) == 1:
            if candidate_speeds[0] < impossible_velocity_threshold:
                candidate_id = candidate_identities[0]
            else:
                candidate_id = 0
        else:
            assert len(candidate_identities) > 1
            if np.count_nonzero(candidate_speeds == np.min(candidate_speeds)) == 1:
                if candidate_speeds[0] < impossible_velocity_threshold:
                    candidate_id = candidate_identities[0]
                else:
                    candidate_id = 0
            else:
                candidate_id = 0

    fragment.identity_corrected_solving_jumps = candidate_id


def get_fragment_with_same_identity(
    list_of_fragments: ListOfFragments,
    fragment: Fragment,
    direction: Literal["to_the_past", "to_the_future"],
) -> tuple[Fragment | None, int]:
    "Get the `neighbour_fragment` with the same identity in a given `direction`"

    if direction == "to_the_past":
        if fragment.identifier == 0:
            return None, 0

        for neighbour in list_of_fragments.fragments[fragment.identifier - 1 :: -1]:
            if (
                neighbour.is_an_individual
                and neighbour.assigned_identities[0] == fragment.assigned_identities[0]
            ):
                number_of_frames_in_direction = (
                    fragment.start_frame - neighbour.end_frame + 1
                )
                return neighbour, number_of_frames_in_direction

    elif direction == "to_the_future":
        for neighbour in list_of_fragments.fragments[fragment.identifier + 1 :]:
            if (
                neighbour.is_an_individual
                and neighbour.assigned_identities[0] == fragment.assigned_identities[0]
            ):
                number_of_frames_in_direction = (
                    neighbour.start_frame - fragment.end_frame + 1
                )
                return neighbour, number_of_frames_in_direction
    else:
        raise ValueError(direction)
    return None, 0


def compute_neighbour_fragments_and_velocities(
    list_of_fragments: ListOfFragments, fragment: Fragment
) -> tuple[Fragment | None, Fragment | None, float | None, float | None]:
    """Computes the fragments with the same identities to the past and to the
    future of a given `fragment` and gives the velocities at the extremes of
    the current `fragment`

    Parameters
    ----------
    session : <Session object>
        Object collecting all the parameters of the video and paths for saving
        and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and
        methods related to them
    fragment : <Fragment object>
        Object collecting all the information for a consecutive set of
        overlapping blobs that are considered to be the same animal

    Returns
    -------
    neighbour_fragment_past : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        past
    neighbour_fragment_future : <Fragment object>
        `Fragment` object with the same identity as the current fragment in the
        future
    velocities_between_fragments : nd.array
        Velocities needed to connect the current fragment to its consecutive
        fragments in the past and in the future.
    """
    neighbour_fragment_past, n_frames_in_past = get_fragment_with_same_identity(
        list_of_fragments, fragment, "to_the_past"
    )
    neighbour_fragment_future, n_frames_in_future = get_fragment_with_same_identity(
        list_of_fragments, fragment, "to_the_future"
    )

    velocity_past = fragment.compute_border_velocity(neighbour_fragment_past)
    velocity_future = fragment.compute_border_velocity(neighbour_fragment_future)

    return (
        neighbour_fragment_past,
        neighbour_fragment_future,
        velocity_past / n_frames_in_past if velocity_past is not None else None,
        velocity_future / n_frames_in_future if velocity_future is not None else None,
    )


def correct_impossible_velocity_jumps_loop(
    session: Session,
    list_of_fragments: ListOfFragments,
    scope: Literal["to_the_past", "to_the_future"],
):
    """Checks whether the velocity needed to join two consecutive fragments with
    the same identity is consistent with the typical velocity of the animals in
    the video (`video.velocity_threshold`). If the velocity is not consistent the
    identity of one of the fragments is reassigned. The check is performed from the
    `video.first_frame_first_global_fragment` to the past or to the future according
    to the `scope`

    Parameters
    ----------
    session : <Session object>
        Object collecting all the parameters of the video and paths for saving
        and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them
    scope : string
        If `scope` = `to_the_past` the check is performed to the past and if
        `scope` = `to_the_future` the check is performed to the future.
    """
    fragments_in_direction = list_of_fragments.get_ordered_list_of_fragments(
        scope, session.first_frame_first_global_fragment
    )
    velocity_threshold = session.velocity_threshold

    # at this point all individual fragments have a valid identity or 0
    for fragment in track(
        fragments_in_direction, f"Correcting impossible velocity jumps {scope}"
    ):
        # TODO This loops is too slow
        if fragment.is_a_crossing or fragment.assigned_identities[0] == 0:
            continue

        (
            neighbour_fragment_past,
            neighbour_fragment_future,
            velocity_to_past,
            velocity_to_future,
        ) = compute_neighbour_fragments_and_velocities(list_of_fragments, fragment)

        if (
            velocity_to_past is not None
            and velocity_to_future is not None
            and velocity_to_past > velocity_threshold
            and velocity_to_future > velocity_threshold
        ):
            assert neighbour_fragment_future is not None
            assert neighbour_fragment_past is not None
            if (
                neighbour_fragment_past.identity_is_fixed
                or neighbour_fragment_future.identity_is_fixed
            ):
                reassign(fragment, list_of_fragments, velocity_threshold)
            else:
                neighbour_fragment_past_past = (
                    neighbour_fragment_past.get_neighbour_fragment(
                        list_of_fragments.fragments, "to_the_past"
                    )
                )
                velocity_past_past = neighbour_fragment_past.compute_border_velocity(
                    neighbour_fragment_past_past
                )

                neighbour_fragment_future_future = (
                    neighbour_fragment_future.get_neighbour_fragment(
                        list_of_fragments.fragments, "to_the_future"
                    )
                )
                velocity_future_future = (
                    neighbour_fragment_future.compute_border_velocity(
                        neighbour_fragment_future_future
                    )
                )

                if (
                    velocity_past_past is not None
                    and velocity_past_past < velocity_threshold
                ) or (
                    velocity_future_future is not None
                    and velocity_future_future < velocity_threshold
                ):
                    reassign(fragment, list_of_fragments, velocity_threshold)
        elif velocity_to_past is not None and velocity_to_past > velocity_threshold:
            assert neighbour_fragment_past is not None
            if neighbour_fragment_past.identity_is_fixed:
                reassign(fragment, list_of_fragments, velocity_threshold)
            else:
                reassign(neighbour_fragment_past, list_of_fragments, velocity_threshold)
        elif velocity_to_future is not None and velocity_to_future > velocity_threshold:
            assert neighbour_fragment_future is not None
            if neighbour_fragment_future.identity_is_fixed:
                reassign(fragment, list_of_fragments, velocity_threshold)
            else:
                reassign(
                    neighbour_fragment_future, list_of_fragments, velocity_threshold
                )


def correct_impossible_velocity_jumps(
    session: Session, list_of_fragments: ListOfFragments
):
    """Corrects the parts of the video where the velocity of any individual is
    higher than a particular velocity threshold given by `video.velocity_threshold`.
    This check is done from the `video.first_frame_first_global_fragment` to the
    past and to the future

    Parameters
    ----------
    session : <Session object>
        Object collecting all the parameters of the video and paths for saving and loading
    list_of_fragments : <ListOfFragments object>
        Object collecting the list of fragments and all the statistics and methods
        related to them

    See Also
    --------
    correct_impossible_velocity_jumps_loop

    """
    correct_impossible_velocity_jumps_loop(
        session, list_of_fragments, scope="to_the_past"
    )
    correct_impossible_velocity_jumps_loop(
        session, list_of_fragments, scope="to_the_future"
    )



from collections.abc import Iterator, Sequence
from functools import cached_property
from statistics import fmean
from typing import Any, Literal

import numpy as np

from . import conf
from .utils import deprecated


class Fragment:
    """A collection of :class:`Blob` s that belong to the same animal or to the same crossing.

    The `Fragment` class represents a sequence of images (or blobs) that are believed to belong to the same animal or
    to a group of animals that are crossing paths. It contains various attributes and methods to manage and analyze
    these sequences, including their positions, velocities, identities, and other properties.
    """

    P1_vector: np.ndarray
    """Numpy array indicating the P1 probability of each of the possible
    identities"""

    coexisting_individual_fragments: Sequence["Fragment"]
    """list of fragment objects representing and individual (i.e.
    not representing a crossing where two or more animals are touching) and
    coexisting (in frame) with self. Doesn't include self."""

    frame_by_frame_velocity: np.ndarray
    """Instant speed (in each frame) of the blob in the fragment"""

    start_position: tuple[float, float]
    """X and Y position of the blob's centroid at the start of the fragment"""

    end_position: tuple[float, float]
    """X and Y position of the blob's centroid at the end of the fragment"""

    images: np.ndarray
    """Indices of Fragment's images in the hdf5 files"""

    loaded_images: np.ndarray
    """Fragment's actual images, in uint8. Intended to be used outside idtrackerai"""

    episodes: np.ndarray
    """Episode where each Fragment images belongs to. It determined the hdf5 file where the image is."""

    groundtruth_identity: int
    """Groundtruth identity assigned externally after validating"""

    identifier: int

    start_frame: int

    end_frame: int

    is_an_individual: bool

    acceptable_for_training: bool | None = None
    """Boolean to indicate that the fragment was identified sufficiently
    well and can in principle be used for training. See also the
    accumulation_manager.py module."""

    temporary_id: int | None = None
    """Integer indicating a temporary identity assigned to the fragment
    during the cascade of training and identification protocols. From 0 to n_animals-1"""

    accumulable: bool | None = None
    """Boolean indicating whether the fragment can be accumulated, i.e. it
    can potentially be used for training."""

    is_in_a_global_fragment: bool = False
    """Indicates whether the fragment is part of a global fragment"""

    certainty: float = 0.0
    """Indicates the certainty of the identity"""

    P2_vector: np.ndarray | None = None
    """Numpy array indicating the P2 probability of each of the possible identities. See also :meth:`compute_P2_vector`"""

    identity: int | None = None
    """Identity assigned to the fragment during the cascade of training
    and identification protocols or during the residual identification. From 1 to n_animals"""

    non_consistent: bool = False
    """Boolean indicating whether the fragment identity is consistent with
    coexisting fragment"""

    ambiguous_identities: np.ndarray | None = None
    """Identities that would be ambiguously assigned during the residual
    identification process. See also the assigner.py module"""

    used_for_training: bool = False
    """Boolean indicating whether the images in the fragment were used to
    train the identification network during the cascade of training and
    identification protocols. See also the accumulation_manager.py module.
    """
    accumulation_step: int | None = None
    """Integer indicating the accumulation step at which the fragment was
    accumulated. See also the accumulation_manager.py module."""

    identities_corrected_closing_gaps: list[int] | None = None
    """Identity of the fragment assigned during the interpolation of the
        gaps produced by the crossing fragments. See also the
        assign_them_all.py module."""

    identity_corrected_solving_jumps: int | None = None
    """Identity of the fragment assigned during the correction of imposible
    (unrealistic) velocity jumps in the trajectories. See also the
    correct_impossible_velocity_jumps.py module."""

    identity_is_fixed: bool = False
    """Boolean indicating whether the identity is fixed and cannot be
    modified during the postprocessing. This attribute is given during
    the residual identification (see assigner.py module)"""

    P1_below_random: bool = False

    accumulated_globally: bool = False
    """Boolean indicating whether the fragment was accumulated in a
    global accumulation step of the cascade of training and identification
    protocols. See also the accumulation_manager.py module."""

    accumulated_partially: bool = False
    """Boolean indicating whether the fragment was accumulated in a
    partial accumulation step of the cascade of training and identification
    protocols. See also the accumulation_manager.py module."""

    forced_crossing: bool = False
    """Indicates if the crossing attribute has been forced by set_individual_with_identity_0_as_crossings()"""

    exclusive_roi: int = -1
    """Exclusive ROI where the fragment belongs to. -1 for disabled exclusive ROIs"""

    zero_identity_assigned_by_P2: bool = False
    zero_identity_assigned_by_exclusive_rois: bool = False

    def __init__(
        self,
        fragment_identifier: int,
        start_frame: int,
        end_frame: int,
        images: list[int] | np.ndarray,
        centroids: list[tuple[float, float]],
        episodes: list[int] | np.ndarray,
        is_an_individual: bool,
        exclusive_roi: int,
    ) -> None:
        self.identifier = fragment_identifier
        self.start_frame = start_frame
        self.end_frame = end_frame
        self.images = np.asarray(images)
        self.episodes = np.asarray(episodes)
        self.is_an_individual = is_an_individual
        self.exclusive_roi = exclusive_roi

        if len(centroids) > 1:
            self.frame_by_frame_velocity = np.sqrt(
                (np.diff(np.asarray(centroids), axis=0) ** 2).sum(axis=1)
            )
        else:
            self.frame_by_frame_velocity = np.array([0])

        self.start_position = centroids[0]
        self.end_position = centroids[-1]

    def __len__(self) -> int:
        return self.n_images

    def __str__(self) -> str:
        return (
            f"<Fragment {self.identifier}, length={len(self)},"
            f" frames={self.start_frame}-{self.end_frame}>"
        )

    @property
    def image_locations(self) -> Iterator[tuple[Any, Any]]:
        return zip(self.images, self.episodes)

    @classmethod
    def from_json(cls, json: dict) -> "Fragment":
        fragment: "Fragment" = cls.__new__(cls)
        fragment.__dict__ = json
        if len(fragment.episodes) == 1:  # v<=5.2.5 decompress
            fragment.episodes = np.full(len(fragment), fragment.episodes[0])
        if (
            len(fragment.episodes) == 2
            and isinstance(fragment.episodes[0], list)
            and isinstance(fragment.episodes[1], list)
        ):  # v>5.2.5 decompress
            fragment.episodes = np.repeat(*fragment.episodes)
        else:
            fragment.episodes = np.asarray(fragment.episodes)

        for key in (
            "images",
            "P1_vector",
            "P2_vector",
            "ambiguous_identities",
            "frame_by_frame_velocity",
        ):
            if key in json:
                setattr(fragment, key, np.asarray(json[key]))
        return fragment

    @property
    def distance_travelled(self) -> float:
        """The distance traveled by the individual in the fragment.
        It is based on the position of the centroids in consecutive images.
        """
        return self.frame_by_frame_velocity.sum()

    def reset(
        self,
        roll_back_to: Literal["fragmentation", "accumulation"],
        number_of_animals: int,
    ) -> None:
        """Reset attributes of the fragment to a specific part of the
        algorithm.

        Parameters
        ----------
        roll_back_to : str
            Reset all the attributes up to the process specified in input.
            'fragmentation', 'accumulation', 'assignment'
        """
        #  This method was mainly used to resume the tracking from different
        # rocessing steps. Currently this function is not active, but this
        #  method might still be useful in the future.
        self.identity_is_fixed = False
        if roll_back_to == "fragmentation":
            self.used_for_training = False
            self.acceptable_for_training = None
            self.temporary_id = None
            self.identity = None
            self.identity_corrected_solving_jumps = None
            self.accumulated_globally = False
            self.accumulated_partially = False
            self.accumulation_step = None
            self.non_consistent = False
            self.certainty = 0.0
            self.P1_vector = np.zeros(number_of_animals)
            self.P1_below_random = False
        elif roll_back_to == "accumulation":
            if not self.used_for_training:
                self.identity = None
                self.identity_corrected_solving_jumps = None
                self.P1_vector = np.zeros(number_of_animals)
            self.ambiguous_identities = None
            self.P2_vector = None
            self.__dict__.pop("certainty_P2", None)
        else:
            raise ValueError(roll_back_to)

    @property
    def is_a_crossing(self) -> bool:
        return not self.is_an_individual

    @property
    def assigned_identities(self) -> list[int] | list[int | None]:
        """Assigned identities (list) by the algorithm considering the
        identification process and the postprocessing steps (correction of
        impossible velocity jumps and interpolation of crossings).

        The fragment can have multiple identities if it is a crossing fragment.
        """
        if self.identities_corrected_closing_gaps is not None:
            return self.identities_corrected_closing_gaps
        if self.identity_corrected_solving_jumps is not None:
            return [self.identity_corrected_solving_jumps]
        return [self.identity]

    @cached_property
    def n_images(self) -> int:
        """Number images (or blobs) in the fragment."""
        return len(self.images)

    @property
    def is_certain(self) -> bool:
        """Whether the fragment is certain enough to be accumulated."""
        return self.certainty >= conf.CERTAINTY_THRESHOLD

    @property
    def has_enough_accumulated_coexisting_fragments(self) -> bool:
        """Whether the fragment has enough coexisting and
        already accumulated fragments (the threshold is half of them).

        This property is used during the partial accumulation. See also the
        accumulation_manager.py module.
        """
        return (
            fmean(
                fragment.used_for_training
                for fragment in self.coexisting_individual_fragments
            )
            >= 0.5
        )

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        state.pop("coexisting_individual_fragments", None)
        state.pop("centroids", None)  # v5.1.3 compatibility
        state.pop("accumulable", None)
        state.pop("n_images", None)  # cached_property
        state.pop("certainty_P2", None)  # cached_property
        return state

    def compute_border_velocity(self, other: "Fragment|None") -> float | None:
        """Velocity necessary to cover the space between two fragments.

        Note that these velocities are divided by the number of frames that
        separate self and other fragment.

        Parameters
        ----------
        other : :class:`Fragment`
            Another fragment

        Returns
        -------
        float
            Returns the speed at which an individual should travel to be
            present in both self and other fragments.

        """
        if other is None:
            return None
        if self.start_frame > other.end_frame:
            centroids = np.asarray([self.start_position, other.end_position])
        else:
            centroids = np.asarray([self.end_position, other.start_position])
        return np.sqrt((np.diff(centroids, axis=0) ** 2).sum(axis=1))[0]

    def coexist_with(self, other: "Fragment") -> bool:
        """Boolean indicating whether the given fragment coexists in time with
        another fragment.

        Parameters
        ----------
        other :  :class:`Fragment`
            A second fragment

        Returns
        -------
        bool
            True if self and other coexist in time in at least one frame.

        """
        return self.start_frame < other.end_frame and self.end_frame > other.start_frame

    def is_inconsistent_with_coexistent_fragments(self, temporary_id) -> bool:
        """Check that the temporary identity assigned to the fragment is
        consistent with respect to the identities already assigned to the
        fragments coexisting (in frame) with it.

        Parameters
        ----------
        temporary_id : int
            Temporary identity assigned to the fragment.

        Returns
        -------
        bool
            True if the identification of self with `temporary_id` does not
            cause any duplication of identities.

        """
        return any(
            coexisting_fragment.temporary_id == temporary_id
            for coexisting_fragment in self.coexisting_individual_fragments
        )

    def set_identification_statistics(
        self,
        predictions_per_img: np.ndarray,
        probabilities_per_img: np.ndarray,
        n_animals: int,
    ) -> None:
        """Computes the statistics necessary for the identification of the
        fragment.

        Parameters
        ----------
        predictions_per_img : numpy array
            Array of shape (number_of_images_in_fragment,)
        probabilities_per_img : numpy array
            Array of shape (number_of_images_in_fragment,)
        n_animals : int
            Number of animals.
        """
        assert self.is_an_individual
        assert len(predictions_per_img) == self.n_images
        assert probabilities_per_img.ndim == predictions_per_img.ndim == 1

        frequencies = np.bincount(predictions_per_img, minlength=n_animals + 1)[1:]
        assert frequencies.sum() == len(self)
        self.set_P1_from_frequencies(frequencies)

        probabilities_per_identity = np.zeros(n_animals)
        for i in np.unique(predictions_per_img):
            probabilities_per_identity[i - 1] = np.median(
                probabilities_per_img[predictions_per_img == i]
            )

        argsort_p1_vector = self.P1_vector.argsort()
        sorted_p1_vector = self.P1_vector[argsort_p1_vector]
        sorted_probabilities = probabilities_per_identity[argsort_p1_vector]
        self.certainty = (
            np.diff((sorted_p1_vector * sorted_probabilities)[-2:])
            / sorted_p1_vector[-2:].sum()
        )[0]

    def assign_identity(
        self, number_of_animals: int, id_to_roi: list[int] | np.ndarray
    ) -> None:
        """Assigns the identity to the fragment by considering the fragments
        coexisting with it.

        If the certainty of the identification is high enough it sets
        the identity of the fragment as fixed and it won't be modified during
        the postprocessing.
        """
        assert self.is_an_individual
        assert not self.used_for_training
        assert not self.identity_is_fixed
        assert self.identity is None
        assert self.P2_vector is not None

        max_P2 = self.P2_vector.max()  # there can be two equal maximums
        possible_identities = np.nonzero(self.P2_vector == max_P2)[0] + 1

        if len(possible_identities) > 1:
            self.identity = 0
            self.zero_identity_assigned_by_P2 = True
            self.ambiguous_identities = possible_identities
            return

        identity: int = possible_identities[0]
        if id_to_roi[identity - 1] != self.exclusive_roi:
            self.identity = 0
            self.zero_identity_assigned_by_exclusive_rois = True
            return

        self.identity = identity
        if (
            max_P2 > conf.FIXED_IDENTITY_THRESHOLD
            and self.n_images >= conf.MIN_N_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION
        ):
            self.identity_is_fixed = True
        self.P1_vector = np.zeros(len(self.P1_vector))
        self.P1_vector[self.identity - 1] = 1.0
        for fragment in self.coexisting_individual_fragments:
            if fragment.identity is None:
                fragment.compute_P2_vector(number_of_animals)

    def compute_P2_vector(self, number_of_animals: int) -> None:
        """Computes the P2_vector of the fragment.

        The flag only_non_identified is to save computational resources when
        Assigning identities after accumulation"""
        self.__dict__.pop("certainty_P2", None)  # clear cached property
        coexisting_P1_vectors = np.asarray(
            [fragment.P1_vector for fragment in self.coexisting_individual_fragments]
        )
        numerator = self.P1_vector * np.prod(1.0 - coexisting_P1_vectors, axis=0)
        denominator = numerator.sum()
        if denominator != 0:
            self.P2_vector = numerator / denominator
        else:
            self.P2_vector = np.zeros(number_of_animals)

    @cached_property
    def certainty_P2(self) -> float:
        """Indicating the certainty of the identity following the P2"""

        if self.P2_vector is None or self.P2_vector.sum() < 0.001:
            return 0.0

        second_max, first_max = np.sort(self.P2_vector)[-2:]

        with np.errstate(divide="ignore", over="ignore"):
            return first_max / second_max

    def set_P1_from_frequencies(self, frequencies: np.ndarray) -> None:
        """Given the frequencies of a individual fragment
        computer the P1 vector.

        P1 is the softmax of the frequencies with base 2 for each identity.
        Numpy array indicating the number of images assigned with each of
        the possible identities
        """
        with np.errstate(over="ignore"):
            self.P1_vector = 1.0 / (
                2.0 ** (frequencies[:, None] - frequencies[None, :])
            ).sum(axis=0)

    def get_neighbour_fragment(
        self,
        fragments: Sequence["Fragment"],
        scope: Literal["to_the_past", "to_the_future"],
    ) -> "Fragment | None":
        """If it exist, gets the fragment in the list of all fragment whose
        identity is the identity assigned to self and whose starting frame is
        the ending frame of self + 1, or ending frame is the starting frame of
        self - 1

        Parameters
        ----------
        fragments : Sequence[Fragment]
            List of all the fragments in the video
        scope : Literal["to_the_past", "to_the_future"]
            If "to_the_future" looks for the consecutive fragment wrt self,
            if "to_the_past" looks for the fragment the precedes self

        Returns
        -------
        Fragment | None
            The neighbor fragment with respect to self in the direction specified by scope if it exists. Otherwise None

        """
        self_index = fragments.index(self)
        if scope == "to_the_past":
            for frag in fragments[self_index - 1 :: -1]:
                if (
                    frag.is_an_individual
                    and frag.assigned_identities[0] == self.assigned_identities[0]
                    and self.start_frame == frag.end_frame
                ):
                    assert len(frag.assigned_identities) == 1
                    return frag

        elif scope == "to_the_future":
            for frag in fragments[self_index + 1 :]:
                if (
                    frag.is_an_individual
                    and frag.assigned_identities[0] == self.assigned_identities[0]
                    and frag.start_frame == self.end_frame
                ):
                    assert len(frag.assigned_identities) == 1
                    return frag
                if frag.start_frame > self.end_frame:
                    # next fragments with have larger and larger start_frame, no chances to find it
                    break

        else:
            raise ValueError(scope)

        return None

    def set_partially_or_globally_accumulated(self, accumulation_strategy) -> None:
        """Sets :attr:`accumulated_globally` and :attr:`accumulated_partially`
        according to `accumulation_strategy`.

        Parameters
        ----------
        accumulation_strategy : Literal["global", "partial"]
            Can be "global" or "partial"
        """
        if accumulation_strategy == "global":
            self.accumulated_globally = True
        elif accumulation_strategy == "partial":
            self.accumulated_partially = True

    @property
    def summary(self) -> Sequence[str]:
        return (
            f"Fragment {self.identifier}",
            f"Frames from {self.start_frame} to {self.end_frame} (length"
            f" {self.end_frame - self.start_frame})",
            ("Individual" if self.is_an_individual else "Crossing")
            + " fragment"
            + (" (forced)" if self.forced_crossing else ""),
            ("Used" if self.used_for_training else "Not used") + " for training",
            ("Acceptable" if self.acceptable_for_training else "Not acceptable")
            + " for training",
            f"Predicted identity: {self.identity}",
            f"Corrected solving jumps: {self.identity_corrected_solving_jumps}",
            f"Corrected solving gaps: {self.identities_corrected_closing_gaps}",
            f"Fixed identity: {self.identity_is_fixed}",
            f"Globally accumulated: {self.accumulated_globally}",
            f"Partially accumulated: {self.accumulated_partially}",
            f"Accumulable: {self.accumulable}",
            f"Accumulated at step {self.accumulation_step}",
            "Non consistent" if self.non_consistent else "Consistent",
            (
                f"Max P1 {self.P1_vector.argmax() + 1} with value {self.P1_vector.max()}"
                if hasattr(self, "P1_vector")
                else "Doesn't have P1 vector"
            ),
            f"Certainty: {self.certainty}",
            f"P1 below random: {self.P1_below_random}",
        )

    # Deprecated properties and methods

    @property
    @deprecated(version="6.0.0")
    def used_for_pretraining(self):
        return False

    @deprecated(
        version="6.0.0", reason="Use :meth:`set_identification_statistics` instead"
    )
    def compute_identification_statistics(self, *args):
        return self.set_identification_statistics(*args)

    @staticmethod
    @deprecated(
        version="6.0.0",
        reason="This method is included in :meth:`set_identification_statistics`",
        action="error",
    )
    def compute_median_softmax(): ...

    @deprecated(
        version="6.0.0",
        reason="This method is included in :meth:`set_identification_statistics`",
        action="error",
    )
    def set_certainty_of_individual_fragment(self): ...

    @property
    @deprecated(version="6.0.0", reason="Use :meth:`summary` instead")
    def properties(self):
        return self.summary



"""The video generator tool is used to generate videos from the tracking
data generated by the idtracker.ai software, see :ref:`video generator`."""

import argparse
import logging
from pathlib import Path

from idtrackerai import IdtrackeraiError, Session
from idtrackerai.utils import wrap_entrypoint

from .general_video import generate_general_video
from .individual_videos import generate_individual_video

__all__ = ["generate_general_video", "generate_individual_video"]


@wrap_entrypoint
def idtrackerai_video_entrypoint() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "session_path",
        type=Path,
        help="Path to the video session created during the tracking session",
        metavar="",
        nargs="?",
    )

    parser.add_argument(
        "--run", action="store_true", help="Generate the video without GUI"
    )

    parser.add_argument(
        "--individual",
        action="store_true",
        help="Generate individual video. Default is a general video",
    )
    parser.add_argument(
        "--gray", action="store_true", help="Draw the original video in grayscale"
    )

    parser.add_argument(
        "--t",
        type=Path,
        help=(
            "Path to the trajectory file, default is "
            "session_dir/trajectories/trajectories*"
        ),
        metavar="",
    )
    parser.add_argument(
        "--tl",
        type=int,
        default=20,
        help=(
            "Trail length, number of points used to draw the individual trajectories"
            " traces in general videos. Default is 20"
        ),
        metavar="",
    )
    parser.add_argument(
        "--s", type=int, default=0, help="Frame where to start the video", metavar=""
    )
    parser.add_argument(
        "--e", type=int, help="Frame where to end the video", metavar=""
    )
    parser.add_argument(
        "--size",
        type=int,
        help=(
            "Size of the squared individual videos. Defaults to the median body length"
            " of the animals."
        ),
        metavar="",
    )
    parser.add_argument(
        "--no-labels",
        action="store_true",
        help="Show centroids in general video without labeled ID's",
    )
    args = parser.parse_args()

    if not args.run:
        run_gui(args.session_path)
        return

    if args.session_path is None:
        raise IdtrackeraiError(
            "A session path is required when using '--run' to generate the video from the terminal."
        )

    try:
        session = Session.load(args.session_path)
    except FileNotFoundError as exc:
        raise IdtrackeraiError() from exc

    if args.no_labels:
        labels = []
    else:
        labels = session.identities_labels or list(
            map(str, range(1, session.n_animals + 1))
        )

    if args.individual:
        if args.no_labels:
            logging.info('Ignoring "--no-labels" flag')

        generate_individual_video(
            session,
            args.t,
            draw_in_gray=args.gray,
            starting_frame=args.s,
            ending_frame=args.e,
            miniframe_size=args.size,
            labels=labels,
        )
    else:
        generate_general_video(
            session,
            args.t,
            draw_in_gray=args.gray,
            centroid_trace_length=args.tl,
            starting_frame=args.s,
            ending_frame=args.e,
            labels=labels,
        )


def run_gui(session: str | Path | None = None) -> None:
    import sys

    from qtpy.QtWidgets import QApplication

    from .video_generator_gui import VideoGeneratorGUI

    app = QApplication(sys.argv)
    window = VideoGeneratorGUI()
    window.show()
    if session:
        window.open(session)
    app.exec()


if __name__ == "__main__":
    idtrackerai_video_entrypoint()



# Each Qt binding is different, so...
# pyright: reportIncompatibleMethodOverride=false
import logging
from importlib import metadata
from pathlib import Path

from qtpy import API_NAME
from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import QSettings, Qt, QThread, QTimer, QUrl
from qtpy.QtGui import (
    QAction,
    QCloseEvent,
    QDesktopServices,
    QDropEvent,
    QGuiApplication,
    QIcon,
)
from qtpy.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLayout,
    QMainWindow,
    QMenu,
    QMessageBox,
    QStyleFactory,
    QWidget,
)

from idtrackerai.utils.telemetry import (
    ComparisonResult,
    check_version,
    get_usage_analytics_state,
    set_usage_analytics_state,
)

from .themes import dark, light


class GUIBase(QMainWindow):

    # in some computers, the tooltip text is white ignoring the palette
    stylesheet: str = (
        "QToolTip { color: black;} QMenu::separator {height: 1px;background: gray; margin-left: 10px; margin-right:10px;}"
    )

    def __init__(self):
        try:
            QT_version = metadata.version(API_NAME)
        except metadata.PackageNotFoundError:
            QT_version = "unknown version"
        logging.info(
            "Initializing %s with %s %s", self.__class__.__name__, API_NAME, QT_version
        )
        if "Fusion" in QStyleFactory.keys():  # noqa SIM118
            QApplication.setStyle("Fusion")
        super().__init__()

        QApplication.setApplicationDisplayName("idtracker.ai")
        QApplication.setApplicationName("idtracker.ai")
        self.setWindowIcon(QIcon(str(Path(__file__).parent / "icon.svg")))
        self.settings = QSettings("idtrackerai", "idtrackerai_GUI")
        logging.debug("Saving GUI settings in %s", self.settings.fileName())

        self.setCentralWidget(QWidget())
        self.centralWidget().setLayout(QHBoxLayout())

        self.documentation_url: str = ""
        """Link to documentation appearing in the menu bar"""

        self.widgets_to_close: list[QWidget] = []
        """Widgets in this list will be called with .close() when closing the app"""

        about_menu = self.menuBar().addMenu("About")
        assert isinstance(about_menu, QMenu)

        doc_action = QAction("Open documentation", self)
        doc_action.setIcon(QIcon.fromTheme("help-browser"))
        about_menu.addAction(doc_action)
        doc_action.triggered.connect(self.open_docs)

        updates = QAction("Check for updates", self)
        updates.setIcon(QIcon.fromTheme("system-software-update"))
        about_menu.addAction(updates)
        updates.triggered.connect(self.check_updates)

        self.updates_action = QAction("Report usage analytics", self)
        self.updates_action.setCheckable(True)
        self.updates_action.setChecked(get_usage_analytics_state())
        about_menu.addAction(self.updates_action)
        self.updates_action.toggled.connect(self.handle_usage_analytics_toggle)

        quit = QAction("Quit app", self)
        quit.setShortcut(Qt.Key.Key_Q)
        quit.setIcon(QIcon.fromTheme("application-exit"))
        quit.triggered.connect(self.close)  # type: ignore

        zoom_in = QAction("Zoom in", self)
        zoom_in.setShortcut("Ctrl++")
        zoom_in.setIcon(QIcon.fromTheme("zoom-in"))
        zoom_in.triggered.connect(lambda: self.change_font_size(1))  # type: ignore

        zoom_out = QAction("Zoom out", self)
        zoom_out.setShortcut("Ctrl+-")
        zoom_out.setIcon(QIcon.fromTheme("zoom-out"))
        zoom_out.triggered.connect(lambda: self.change_font_size(-1))  # type: ignore

        self.themeAction = QAction("Dark theme", self)
        self.themeAction.toggled.connect(self.change_theme)
        self.themeAction.setCheckable(True)
        self.change_theme(False)

        view_menu = self.menuBar().addMenu("View")
        assert view_menu is not None
        view_menu.addAction(quit)
        view_menu.addSeparator()
        view_menu.addAction(zoom_in)
        view_menu.addAction(zoom_out)
        view_menu.addAction(self.themeAction)

        dark_theme = self.settings.value("dark_theme", False, type=bool)
        font_size = self.settings.value("fontsize", self.font().pointSize(), type=int)

        self.themeAction.setChecked(dark_theme)
        font = self.font()
        font.setPointSize(font_size)
        self.setFont(font)
        QApplication.setFont(font)
        self.change_theme(dark_theme)

        self.auto_check_updates = AutoCheckUpdatesThread()
        self.auto_check_updates.out_of_date.connect(
            lambda msg: QMessageBox.about(self, "Check for updates", msg)
        )
        QTimer.singleShot(100, self.auto_check_updates.start)
        self.center_window()
        self.setAcceptDrops(True)

    def change_font_size(self, change: int) -> None:
        font = self.font()
        font.setPointSize(min(max(font.pointSize() + change, 5), 20))
        self.setFont(font)
        QApplication.setFont(font)
        # This has to be here so that the font size change takes place
        self.setStyleSheet(self.stylesheet)

    def check_updates(self):
        if hasattr(check_version, "cache_clear"):
            check_version.cache_clear()  # Clear check_version cache
        kind, message = check_version()
        if kind == ComparisonResult.ERROR:
            QMessageBox.critical(self, "Error checking for updates", message)
        else:
            QMessageBox.about(self, "Check for updates", message)

    def open_docs(self):
        QDesktopServices.openUrl(QUrl(self.documentation_url))

    def center_window(self):
        settings_key = f"{self.__class__.__name__}_window_geometry"
        if self.settings.contains(settings_key):
            self.restoreGeometry(self.settings.value(settings_key))
            return

        w, h = 1000, 800
        try:
            cp = (
                QGuiApplication.screenAt(self.cursor().pos())
                .availableGeometry()
                .center()
            )
        except AttributeError:
            # in Fedora QGuiApplication.screenAt(self.cursor().pos()) is None
            cp = QGuiApplication.primaryScreen().availableGeometry().center()

        self.setGeometry(cp.x() - w // 2, cp.y() - h // 2, w, h)

    def change_theme(self, dark_theme: bool):
        if dark_theme:
            QApplication.setPalette(dark)
        else:
            QApplication.setPalette(light)

        self.setStyleSheet(self.stylesheet)

    def closeEvent(self, event: QCloseEvent):
        self.settings.setValue("dark_theme", self.themeAction.isChecked())
        self.settings.setValue("fontsize", self.font().pointSize())
        self.settings.setValue(
            f"{self.__class__.__name__}_window_geometry", self.saveGeometry()
        )
        for widget_to_close in self.widgets_to_close:
            widget_to_close.close()
        super().closeEvent(event)

    def dragEnterEvent(self, event: QDropEvent) -> None:
        # accept Drag&Drop if it's about files
        data = event.mimeData()
        if data is not None and data.hasUrls():
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent) -> None:
        data = event.mimeData()
        if data is not None and data.hasUrls():
            # send the drop files to self.open_widget
            urls = [url.toLocalFile() for url in data.urls()]
            QTimer.singleShot(0, lambda: self.manageDropedPaths(urls))
            event.accept()

    def manageDropedPaths(self, paths: list[str]) -> None:
        raise NotImplementedError

    def clearFocus(self):
        focused_widged = self.focusWidget()
        if focused_widged:
            focused_widged.clearFocus()

    def mousePressEvent(self, event):
        self.clearFocus()
        super().mousePressEvent(event)

    @staticmethod
    def get_list_of_widgets(layout: QLayout) -> list[QWidget]:
        widgets = []
        layouts = [layout]
        while layouts:
            element = layouts.pop()
            if hasattr(element.widget(), "setEnabled"):
                widgets.append(element.widget())
            else:
                layouts += [element.itemAt(i) for i in range(element.count())]
        return widgets

    def handle_usage_analytics_toggle(self, checked: bool):
        if not checked:
            reply = QMessageBox.question(
                self,
                "Disable Usage Analytics",
                "Usage analytics are anonymized and help us improve idtracker.ai. "
                "It only collects general information such as the operating system, "
                "the version of idtracker.ai being used, and the command used to run "
                "it.\n\nAre you sure you want to disable reporting usage analytics?",
            )
            if reply != QMessageBox.StandardButton.Yes:
                self.updates_action.setChecked(True)
                return
        set_usage_analytics_state(checked)


class AutoCheckUpdatesThread(QThread):
    out_of_date = Signal(str)

    def run(self):
        kind, message = check_version()
        if kind in (ComparisonResult.MAJOR_UPDATE, ComparisonResult.MINOR_UPDATE):
            self.out_of_date.emit(message)



from collections.abc import Iterator, Sequence
from typing import Literal

from . import Fragment, IdtrackeraiError
from .utils import deprecated


class GlobalFragment:
    """Represents a collection of N_animals coexisting :class:`Fragment` s."""

    duplicated_identities: set

    fragments_identifiers: Sequence[int]

    fragments: Sequence[Fragment]

    accumulation_step: int | None = None
    """Integer indicating the accumulation step at which the global fragment was globally accumulated."""

    def __init__(self, fragments: Sequence[Fragment]) -> None:
        self.fragments_identifiers = tuple(frag.identifier for frag in fragments)
        self.fragments = fragments

        for fragment in self:
            fragment.is_in_a_global_fragment = True

    @property
    def minimum_distance_travelled(self) -> float:
        return min(fragment.distance_travelled for fragment in self)

    @property
    def first_frame_of_the_core(self) -> int:
        return max(frag.start_frame for frag in self)

    @property
    def min_n_images_per_fragment(self) -> int:
        return min(fragment.n_images for fragment in self)

    def __iter__(self) -> Iterator[Fragment]:
        try:
            return iter(self.fragments)
        except AttributeError as exc:
            raise IdtrackeraiError(
                "Global Fragment has been loaded without access to Fragments"
            ) from exc

    @classmethod
    def from_json(
        cls, data: dict, fragments: list[Fragment] | None
    ) -> "GlobalFragment":
        global_fragment = cls.__new__(cls)
        if "individual_fragments_identifiers" in data:
            data["fragments_identifiers"] = data.pop("individual_fragments_identifiers")
        data.pop("first_frame_of_the_core", None)  # it wasn't a property before 6.0.0
        global_fragment.__dict__.update(data)
        if "duplicated_identities" in data:
            global_fragment.duplicated_identities = set(data["duplicated_identities"])

        if fragments is not None:
            global_fragment.fragments = [
                fragments[frag_id] for frag_id in global_fragment.fragments_identifiers
            ]

        return global_fragment

    @property
    def used_for_training(self) -> bool:
        """Boolean indicating if all the fragments in the global fragment
        have been used for training the identification network"""
        return all(fragment.used_for_training for fragment in self)

    def is_unique(self, number_of_animals: int) -> bool:
        """Boolean indicating that the global fragment has unique
        identities, i.e. it does not have duplications."""
        return {fragment.temporary_id for fragment in self} == set(
            range(number_of_animals)
        )

    @property
    def is_partially_unique(self) -> bool:
        """Boolean indicating that a subset of the fragments in the global
        fragment have unique identities"""

        identities_acceptable_for_training = [
            fragment.temporary_id
            for fragment in self
            if fragment.acceptable_for_training
        ]
        self.duplicated_identities = {
            x
            for x in identities_acceptable_for_training
            if identities_acceptable_for_training.count(x) > 1
        }
        return len(self.duplicated_identities) == 0

    def acceptable_for_training(
        self, accumulation_strategy: Literal["global", "partial"]
    ) -> bool:
        """Returns True if self is acceptable for training"""

        return (all if accumulation_strategy == "global" else any)(
            fragment.acceptable_for_training for fragment in self
        )

    @property
    def total_number_of_images(self) -> int:
        """Gets the total number of images in self"""
        return sum(fragment.n_images for fragment in self)

    @deprecated(version="6.0.0", action="error")
    def get_images_and_labels(self): ...



import logging
from shutil import copyfileobj

from idtrackerai import ListOfBlobs, ListOfFragments, ListOfGlobalFragments, Session
from idtrackerai.utils import TMP_LOG_FILE, Timer

from .animals_detection import animals_detection_API
from .crossings_detection import crossings_detection_API
from .fragmentation import fragmentation_API
from .postprocess import trajectories_API
from .tracker import tracker_API


class RunIdTrackerAi:
    session: Session
    list_of_blobs: ListOfBlobs
    list_of_fragments: ListOfFragments
    list_of_global_fragments: ListOfGlobalFragments

    def __init__(self, session: Session):
        self.session = session

    def track_video(self) -> bool:
        try:
            general_timer = Timer("Tracking session")
            general_timer.start()

            self.session.prepare_tracking()
            self.session.timers["Tracking session"] = general_timer

            self.save()
            with self.session.new_timer("Animal detection"):
                self.list_of_blobs = animals_detection_API(self.session)

            self.save()

            with self.session.new_timer("Crossing detection"):
                crossings_detection_API(self.session, self.list_of_blobs)

            self.save()

            with self.session.new_timer("Fragmentation"):
                self.list_of_fragments, self.list_of_global_fragments = (
                    fragmentation_API(self.session, self.list_of_blobs)
                )

            self.save()

            with self.session.new_timer("Tracking"):
                identifier_model = tracker_API(
                    self.session,
                    self.list_of_blobs,
                    self.list_of_fragments,
                    self.list_of_global_fragments,
                )

            self.save()

            with self.session.new_timer("Trajectories creation"):
                trajectories_API(
                    self.session,
                    self.list_of_blobs,
                    self.list_of_fragments,
                    identifier_model,
                )

            self.session.timers["Tracking session"].finish()
            self.save()

            if self.session.track_wo_identities:
                logging.info(
                    "Tracked without identities, no estimated accuracy available."
                )
            else:
                logging.info(
                    f"Estimated accuracy: {self.session.estimated_accuracy:.4%}"
                )

            self.session.delete_data()
            self.session.compress_data()
            logging.info("[green]Success", extra={"markup": True})
            success = True

        except (Exception, KeyboardInterrupt) as error:

            if (
                hasattr(self, "session")
                and hasattr(self.session, "session_folder")
                and self.session.session_folder.is_dir()
            ):
                # we add the path where we would like to have a copy of the log
                # TODO when Python >= 3.11 use Exception.add_note()
                error.log_path = self.session.session_folder / "idtrackerai.log"  # type: ignore
            raise error

        if (
            hasattr(self, "session")
            and hasattr(self.session, "session_folder")
            and self.session.session_folder.is_dir()
        ):
            log_copy_path = self.session.session_folder / "idtrackerai.log"
            TMP_LOG_FILE.flush()
            TMP_LOG_FILE.seek(0)
            with open(log_copy_path, "w") as file:
                copyfileobj(TMP_LOG_FILE, file)
            logging.info(f"Log file copied to {log_copy_path}")
        return success

    def save(self):
        try:
            if hasattr(self, "session") and hasattr(self.session, "session_folder"):
                self.session.save()
            if hasattr(self, "list_of_blobs"):
                self.list_of_blobs.save(self.session.blobs_path)
            if hasattr(self, "list_of_fragments"):
                self.list_of_fragments.save(self.session.fragments_path)
            if hasattr(self, "list_of_global_fragments"):
                self.list_of_global_fragments.save(self.session.global_fragments_path)
        except Exception as exc:
            logging.error("Error while saving data: %s", exc)



import logging

import numpy as np

from idtrackerai import Blob, conf
from idtrackerai.utils import track


def get_train_validation_and_eval_blobs(
    blobs_in_video: list[list[Blob]],
    number_of_animals: int,
    ratio_val: float = 0.1,
    max_images_per_class: int = conf.MAX_IMAGES_PER_CLASS_CROSSING_DETECTOR,
) -> tuple[np.ndarray, np.ndarray, list[float], np.ndarray, np.ndarray]:
    """Given a list of blobs return 2 dictionaries (training_blobs, validation_blobs),
    and a list (toassign_blobs).
    """
    logging.info("Get list of blobs for training, validation and eval")

    individuals = []
    crossings = []
    for blobs_in_frame in track(blobs_in_video, "First individual/crossing assignment"):
        in_a_global_fragment_core = len(blobs_in_frame) == number_of_animals
        for blob in blobs_in_frame:
            if in_a_global_fragment_core or blob.is_a_sure_individual():
                blob.used_for_training_crossings = True
                blob.is_an_individual = True
                individuals.append((blob.id_image_index, blob.episode))
            elif blob.is_a_sure_crossing():
                blob.used_for_training_crossings = True
                blob.is_an_individual = False
                crossings.append((blob.id_image_index, blob.episode))
            # unknown blobs are recognized by not having the attribute `is_an_individual`

    individuals = np.asarray(individuals) if individuals else np.empty((0, 2))
    crossings = np.asarray(crossings) if crossings else np.empty((0, 2))

    # clear no longer useful cached properties
    for blobs_in_frame in blobs_in_video:
        for blob in blobs_in_frame:
            blob.__dict__.pop("has_a_next_crossing", None)
            blob.__dict__.pop("has_a_previous_crossing", None)
            blob.__dict__.pop("has_multiple_next", None)
            blob.__dict__.pop("has_multiple_previous", None)

    logging.debug(
        f"{len(individuals)} sure individuals and {len(crossings)} sure crossings"
    )

    # Shuffle
    rng = np.random.default_rng()
    rng.shuffle(individuals)
    rng.shuffle(crossings)

    # Limit the number of images
    crossings = crossings[:max_images_per_class]
    individuals = individuals[:max_images_per_class]

    # Split train and val
    n_blobs_crossings = len(crossings)
    n_blobs_individuals = len(individuals)
    n_individual_blobs_val = int(n_blobs_individuals * ratio_val)
    n_crossing_blobs_val = int(n_blobs_crossings * ratio_val)

    val_individual = individuals[:n_individual_blobs_val]
    val_crossing = crossings[:n_crossing_blobs_val]
    train_individual = individuals[n_individual_blobs_val:]
    train_crossing = crossings[n_crossing_blobs_val:]

    val_images = np.concatenate((val_individual, val_crossing))
    val_labels = np.concatenate(
        (np.zeros(len(val_individual), np.int64), np.ones(len(val_crossing), np.int64))
    )

    train_images = np.concatenate((train_individual, train_crossing))
    train_labels = np.concatenate(
        (
            np.zeros(len(train_individual), np.int64),
            np.ones(len(train_crossing), np.int64),
        )
    )

    logging.info(
        f"{len(train_individual)} individual and "
        f"{len(train_crossing)} crossing blobs for training\n"
        f"{len(val_individual)} individual and "
        f"{len(val_crossing)} crossing blobs for validation"
    )

    if n_blobs_crossings + n_blobs_individuals:
        ratio_crossings = n_blobs_crossings / (n_blobs_crossings + n_blobs_individuals)
        train_weights = [ratio_crossings, 1 - ratio_crossings]
    else:
        # there are no sure crossings, nor sure individuals
        # We are not gonna train crossings but this function will respect its type hints
        train_weights = [np.nan, np.nan]

    return train_images, train_labels, train_weights, val_images, val_labels



import json
import logging
import pickle
from collections.abc import Iterable, Iterator
from itertools import pairwise
from pathlib import Path
from typing import Any

import numpy as np

from . import Blob, Fragment, GlobalFragment, conf
from .utils import deprecated, resolve_path


class GlobalFragmentsEncoder(json.JSONEncoder):
    """Json encoder to serialize Global Fragments with styled indentation"""

    def default(self, o) -> Any:
        match o:
            case set():
                return list(o)
            case GlobalFragment():
                serial = o.__dict__.copy()
                serial.pop("fragments", None)  # remove connections

                serial["fragments_identifiers"] = (  # without indentation
                    f"NotString{json.dumps(o.fragments_identifiers)}"
                )

                return serial
            case np.integer():
                return int(o)
            case np.floating():
                return float(o)
            case _:
                return super().default(o)

    def iterencode(self, o: Any, _one_shot: bool = False) -> Iterator[str]:
        return (
            encoded[10:-1] if encoded.startswith('"NotString') else encoded
            for encoded in super().iterencode(o, _one_shot)
        )


class ListOfGlobalFragments:
    """Contains all the instances of the class :class:`GlobalFragment`.

    It contains methods to retrieve information from these global fragments
    and to update their attributes.
    These methods are mainly used during the cascade of training and
    identification protocols.

    Parameters
    ----------
    global_fragments : list
        List of instances of :class:`GlobalFragment`.
    """

    non_accumulable_global_fragments: list[GlobalFragment]
    """List of global fragments which are NOT candidate for accumulation"""

    global_fragments: list[GlobalFragment]
    """List of global fragments which are candidate for accumulation"""

    def __init__(self, global_fragments: Iterable[GlobalFragment]):
        self.global_fragments = []
        self.non_accumulable_global_fragments = []

        for global_fragment in global_fragments:
            if (
                global_fragment.min_n_images_per_fragment
                >= conf.MIN_N_FRAMES_TO_BE_A_CANDIDATE_FOR_ACCUMULATION
            ):
                self.global_fragments.append(global_fragment)
            else:
                self.non_accumulable_global_fragments.append(global_fragment)

        logging.info(
            "Total number of Global Fragments: %d",
            len(self.non_accumulable_global_fragments) + len(self.global_fragments),
        )
        logging.info(
            "Of which %d are long enough to be accumulated", len(self.global_fragments)
        )

    @classmethod
    def from_fragments(
        cls,
        blobs_in_video: list[list[Blob]],
        fragments: list[Fragment],
        num_animals: int,
    ) -> "ListOfGlobalFragments":
        """Creates the list of instances of the class :class:`GlobalFragment`.

        Parameters
        ----------
        blobs_in_video : list[list[Blob]]
            All blobs in video
        fragments : list[Fragment]
            All Fragments in video
        num_animals : int
            Number of animals to be tracked as indicated by the user.

        Returns
        -------
        ListOfGlobalFragments
            The list of all Global Fragments
        """
        is_global_fragment_core = [False] + [
            get_global_fragment_core(blobs_in_frame, blobs_in_past, num_animals)
            for blobs_in_past, blobs_in_frame in pairwise(blobs_in_video)
        ]

        global_fragments_first_frames = [
            frame
            for frame, (was_core, is_core) in enumerate(
                pairwise(is_global_fragment_core), 1
            )
            if is_core and not was_core
        ]

        return cls(
            GlobalFragment(
                [fragments[blob.fragment_identifier] for blob in blobs_in_video[i]]
            )
            for i in global_fragments_first_frames
        )

    def __len__(self) -> int:
        return len(self.global_fragments)

    def __iter__(self) -> Iterator[GlobalFragment]:
        return iter(self.global_fragments)

    def sort_by_distance_travelled(self):
        self.global_fragments.sort(
            key=lambda x: x.minimum_distance_travelled, reverse=True
        )

    def sort_by_distance_to_the_frame(self, frame_number: int):
        """Sorts the global fragments with respect to their distance from the
        first global fragment chose for accumulation.
        """
        self.global_fragments.sort(
            key=lambda x: abs(x.first_frame_of_the_core - frame_number)
        )

    def save(self, path: Path | str):
        """Saves an instance of the class.

        Before saving the instances of fragments associated to every global
        fragment are removed and reset them after saving. This
        prevents problems when pickling objects inside of objects.

        Parameters
        ----------
        path : Path | str
            Path where the object will be stored
        """
        path = resolve_path(path)
        logging.info(f"Saving ListOfGlobalFragments at {path}", stacklevel=2)
        path.parent.mkdir(exist_ok=True)

        json.dump(self.__dict__, path.open("w"), cls=GlobalFragmentsEncoder, indent=4)

    @classmethod
    def load(
        cls, path: Path | str, fragments: list[Fragment] | None = None
    ) -> "ListOfGlobalFragments":
        """Loads an instance of the class saved with :meth:`save`

        Parameters
        ----------

        path : str
            Path where the object to be loaded is stored.
        fragments : list
            List of all the instances of the class :class:`Fragment`
            in the video.
        """
        path = resolve_path(path)
        logging.info(f"Loading ListOfGlobalFragments from {path}", stacklevel=2)

        if not path.is_file():  # <=5.1.3 compatibility
            if not path.with_suffix(".pickle").is_file():
                raise FileNotFoundError(path)
            pickle.load(path.with_suffix(".pickle").open("rb")).save(path)

        list_of_global_fragments = cls.__new__(cls)
        json_data = json.load(path.open("r"))

        list_of_global_fragments.global_fragments = [
            GlobalFragment.from_json(g_frag_data, fragments)
            for g_frag_data in json_data["global_fragments"]
        ]

        list_of_global_fragments.non_accumulable_global_fragments = [
            GlobalFragment.from_json(g_frag_data, fragments)
            for g_frag_data in json_data["non_accumulable_global_fragments"]
        ]

        return list_of_global_fragments

    @deprecated(
        version="6.0.0",
        reason="Use `len(ListOfGlobalFragments.global_fragments) == 1` instead",
    )
    def single_global_fragment(self) -> bool:
        """Returns True if there is only one global fragment in the list."""
        return len(self.global_fragments) == 1

    @deprecated(
        version="6.0.0",
        reason="Use `len(ListOfGlobalFragments.global_fragments) == 0` instead",
    )
    def no_global_fragment(self) -> bool:
        """Returns True if there are no global fragments in the list."""
        return len(self.global_fragments) == 0


def get_global_fragment_core(
    blobs_in_frame: list[Blob], blobs_in_past_frame: list[Blob], n_animals: int
) -> bool:
    """Return True if the frame belongs to a Global Fragment core.

    The last condition in the return line is there because we will look for
    [..., False, True, ...] patterns to get the first frame of the Global Fragment core.
    So we set the frame i to False if the previous frame is in a different
    Global Fragment (which means was_clean_frame and same_fragment_identifiers = False).
    """
    if n_animals == 0:  # unknown number of animals
        return False
    is_clean_frame = len(blobs_in_frame) == n_animals and all(
        b.is_an_individual for b in blobs_in_frame
    )
    was_clean_frame = len(blobs_in_past_frame) == n_animals and all(
        b.is_an_individual for b in blobs_in_past_frame
    )

    same_fragment_identifiers = {b.fragment_identifier for b in blobs_in_frame} == {
        b.fragment_identifier for b in blobs_in_past_frame
    }
    return is_clean_frame and (same_fragment_identifiers or not was_clean_frame)



"""idmatcher.ai matches identities from two different tracked
sessions, see :ref:`idmatcher.ai`."""

import json
import logging
from argparse import ArgumentParser
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import linear_sum_assignment

from idtrackerai import ListOfFragments, Session
from idtrackerai.base.network import (
    IdentifierBase,
    get_predictions,
    load_identifier_model,
)
from idtrackerai.utils import create_dir, resolve_path, wrap_entrypoint

plt.rcParams["font.family"] = "STIXgeneral"


@wrap_entrypoint
def idmatcherai_entrypoint() -> None:
    for handler in logging.root.handlers:
        handler.setLevel(logging.INFO)

    parser = ArgumentParser()
    parser.add_argument(
        "sessions",
        help="path to the session folder with the results from the first video",
        type=path,
        nargs="+",
    )
    args = parser.parse_args()

    idmatcherai(*args.sessions)


def idmatcherai(*folders: Path | str | Session) -> None:
    """idmatcherai script, called by the command ``idmatcherai``.


    .. seealso::
        Documentation for :ref:`idmatcher.ai`

    Parameters
    ----------
    *folders : Path | str | Session
        Sequence of Sessions or session paths to match with idmatcherai. With two elements, the matching is performed between them. With more than two elements, the matching will be computed between the first and the second, the first and the third, the first and the fourth, and so on. If the paths to the Sessions are given, :meth:`idtrackerai.Session.load` will be used to load them.
    """
    logging.info(
        "Matching sessions:\n    "
        + "\n    ".join(map(str, folders[1:]))
        + f"\nwith {folders[0]}"
    )
    master_session = folders[0]
    if not isinstance(master_session, Session):
        master_session = Session.load(master_session)

    master_fragments = ListOfFragments.load(
        master_session.fragments_path, reconnect=False
    )

    for matching_session in folders[1:]:
        if not isinstance(matching_session, Session):
            matching_session = Session.load(matching_session)

        logging.info("\nMatching %s", matching_session)
        if matching_session.n_animals != master_session.n_animals:
            logging.warning(
                "Different number of animals between\n   "
                f" {matching_session} ({matching_session.n_animals})"
                " and\n   "
                f" {master_session} ({master_session.n_animals})"
            )

        if matching_session.version != master_session.version:
            logging.warning(
                "Different idtracker.ai versions between\n    "
                f"{matching_session} {matching_session.version}"
                " and\n    "
                f"{master_session} {master_session.version}\n"
                "This can cause poor matchings"
            )

        if matching_session.id_image_size != master_session.id_image_size:
            logging.error(
                "Different identification image size between\n    "
                f"{matching_session} {matching_session.id_image_size}"
                " and\n    "
                f"{master_session} {master_session.id_image_size}\n"
                "Check how to define a fixed identification image size in"
                " https://idtracker.ai/latest/user_guide/usage.html#knowledge-and-identity-transfer"
            )
            continue

        results_path = (
            matching_session.session_folder
            / "matching_results"
            / (master_session.session_folder.name)
        )
        create_dir(results_path)
        create_dir(results_path / "csv")
        create_dir(results_path / "png")

        direct_matches = match(
            ListOfFragments.load(matching_session.fragments_path, reconnect=False),
            matching_session.id_images_file_paths,
            master_session.accumulation_folder,
        )
        save_matrix(
            direct_matches,
            results_path,
            "direct_matches",
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )

        indirect_matches = match(
            master_fragments,
            master_session.id_images_file_paths,
            matching_session.accumulation_folder,
        ).T
        save_matrix(
            indirect_matches,
            results_path,
            "indirect_matches",
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )

        joined_matches = direct_matches + indirect_matches
        save_matrix(
            joined_matches,
            results_path,
            "joined_matches",
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )

        logging.info(f"{master_session.n_animals} animals in {master_session}")
        logging.info(f"{matching_session.n_animals} animals in {matching_session}")

        if matching_session.n_animals == master_session.n_animals:
            logging.info("Maximizing direct and indirect matches")
            matrix_to_optimize = joined_matches
        elif matching_session.n_animals > master_session.n_animals:
            logging.info("Maximizing indirect matches only")
            matrix_to_optimize = indirect_matches
        elif matching_session.n_animals < master_session.n_animals:
            logging.info("Maximizing direct matches only")
            matrix_to_optimize = direct_matches
        else:
            raise RuntimeError(matching_session.n_animals, master_session.n_animals)

        logging.info("Assigning identities")
        assigned_ids, assignments = linear_sum_assignment(
            matrix_to_optimize, maximize=True
        )

        agreement = (
            joined_matches[assigned_ids, assignments].sum()
            / joined_matches[assigned_ids].sum()
        )

        direct_scores = [
            score_row(direct_matches[assigned_id], assignment)
            for assigned_id, assignment in zip(assigned_ids, assignments)
        ]

        indirect_scores = [
            score_row(indirect_matches[:, assignment], assigned_id)
            for assigned_id, assignment in zip(assigned_ids, assignments)
        ]

        with (results_path / "assignments.csv").open("w", encoding="utf_8") as file:
            file.write("identity, assignment, direct score, indirect score\n")
            for i, assigned_id in enumerate(assigned_ids):
                file.write(
                    f"{assigned_id + 1:8d}, {assignments[i] + 1:10d},"
                    f" {direct_scores[i]:12.4f}, {indirect_scores[i]:14.4f}\n"
                )

        mean_direct_score = float(np.mean(direct_scores))
        mean_indirect_score = float(np.mean(indirect_scores))
        mean_score = (mean_direct_score + mean_indirect_score) / 2
        save_matrix(
            direct_matches,
            results_path,
            "direct_matches",
            (assignments, assigned_ids, mean_direct_score),
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )
        save_matrix(
            indirect_matches,
            results_path,
            "indirect_matches",
            (assignments, assigned_ids, mean_indirect_score),
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )
        save_matrix(
            joined_matches,
            results_path,
            "joined_matches",
            (assignments, assigned_ids, mean_score),
            xlabel=master_session.session_folder.name,
            ylabel=matching_session.session_folder.name,
        )
        logging.info("Results in %s", results_path)
        logging.info(
            "Scores:\n"
            f"    Matching agreement: {agreement:.2%}\n"
            f"    Mean direct score: {mean_direct_score:.2%}\n"
            f"    Mean indirect score: {mean_indirect_score:.2%}\n"
            f"    Total mean score: {mean_score:.2%}"
        )


def match(
    fragments: ListOfFragments, id_images_paths: list[Path], model_path: Path
) -> np.ndarray:

    if not model_path.exists() and model_path.with_name("accumulation_0").exists():
        # backward compatibility
        model_path = model_path.with_name("accumulation_0")

    logging.info(
        "Matching images from %s with model from %s",
        id_images_paths[0].parent,
        model_path,
    )

    image_locations_list: list[tuple[int, int]] = []
    labels_list: list[int] = []
    for fragment in fragments.individual_fragments:
        if fragment.identity not in (None, 0):
            image_locations_list += fragment.image_locations
            labels_list += [fragment.identity] * len(fragment)
    labels = np.asarray(labels_list)
    image_locations = np.asarray(image_locations_list)
    image_locations_list.clear()  # clear the list to save memory
    labels_list.clear()  # clear the list to save memory

    max_images = 10000 * fragments.n_animals
    n_images = len(image_locations)
    if n_images > max_images:
        logging.info(
            "Too many images (%d), randomly sampling 10000 images per animal", n_images
        )
        sample = np.random.choice(n_images, max_images, replace=False)
        image_locations = image_locations[sample]
        labels = labels[sample]

    # we sort the images to speedup the loading process
    order = np.lexsort(image_locations.T)
    image_locations = image_locations[order]
    labels = labels[order]

    model, model_n_classes = load_identification_model(model_path)
    all_predictions = get_predictions(model, image_locations, id_images_paths)[0]

    set_of_labels = np.unique(labels)

    n_img_ids = len(set_of_labels)
    """number of labels in the images to be assigned by the model"""

    n_model_ids = model_n_classes
    """number of classes in the model"""

    confusion_matrix = np.zeros((n_img_ids, n_model_ids), int)

    for identity in set_of_labels:
        predictions = all_predictions[labels == identity]
        confusion_matrix[identity - 1] = np.bincount(
            predictions, minlength=n_model_ids + 1
        )[1:]
    return confusion_matrix


def load_identification_model(model_folder: Path) -> tuple[IdentifierBase, int]:
    params_path = model_folder / "model_params.json"
    if params_path.is_file():
        params = json.loads(params_path.read_text())
    elif params_path.with_suffix(".npy").is_file():
        params = np.load(params_path.with_suffix(".npy"), allow_pickle=True).item()
    else:
        raise FileNotFoundError(params_path)

    n_classes = (  # 5.1.6 compatibility
        params["n_classes"] if "n_classes" in params else params["number_of_classes"]
    )

    identification_model = load_identifier_model(model_folder, params["image_size"])
    return identification_model, n_classes


def score_row(row: np.ndarray, assigned) -> float:
    major_indices = row.argsort()[::-1]
    major_value = row[major_indices[0]]
    if major_value == 0:
        return 0
    for major_competitor in major_indices:
        if major_competitor != assigned:
            return float((row[assigned] - row[major_competitor]) / major_value)
    raise ValueError


def path(value: str) -> Path:
    return_path = resolve_path(value)
    if not return_path.exists():
        raise ValueError()
    return return_path


def save_matrix(
    mat: np.ndarray,
    dir: Path,
    name: str,
    assign: tuple[np.ndarray, np.ndarray, float] | None = None,
    xlabel: str = "",
    ylabel: str = "",
) -> None:
    np.savetxt(
        (dir / "csv" / name).with_suffix(".csv"),
        mat,
        "%5d" if mat.shape[1] < 20 else "%d",
        delimiter=",",
    )
    fig, ax = plt.subplots(figsize=(6, 5), dpi=200)
    im = ax.imshow(
        mat,
        interpolation="none",
        extent=(+0.5, mat.shape[1] + 0.5, mat.shape[0] + 0.5, +0.5),
    )

    # avoid confusions
    if name.startswith("direct"):
        xlabel = "Predictions by model from " + xlabel
        ylabel = "Images from " + ylabel
    elif name.startswith("indirect"):
        xlabel = "Images from " + xlabel
        ylabel = "Predictions by model from " + ylabel

    ax.set(
        title=name.replace("_", " ").capitalize(),
        xlabel=xlabel,
        ylabel=ylabel,
        aspect="auto",
    )
    if assign is not None:
        ax.plot(assign[0] + 1, assign[1] + 1, "rx", ms=8, label="Assignment")
        ax.legend()
        direction = name.split("_")[0]
        ax.set_title(
            ax.get_title() + f" | Assignment {direction} score: {assign[2]:.2%}"
        )

    # show grid
    ax.set_xticks(np.arange(1.5, mat.shape[1]), minor=True)
    ax.set_yticks(np.arange(1.5, mat.shape[0]), minor=True)
    ax.grid(which="minor", color="w")
    ax.tick_params(which="minor", bottom=False, left=False)

    fig.colorbar(im).set_label("Number of matches")
    fig.tight_layout(pad=0.8)
    fig.savefig(str((dir / "png" / name).with_suffix(".png")))
    plt.close()



from collections.abc import Sequence
from pathlib import Path

import numpy as np
from qtpy.QtCore import Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtCore import Qt, QThread, QTimer
from qtpy.QtGui import QImage, QPainter, QPixmap
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QMessageBox,
    QProgressDialog,
    QSizePolicy,
    QToolButton,
    QWidget,
)

from idtrackerai.base.animals_detection import (
    generate_background_from_frame_stack,
    generate_frame_stack,
    load_custom_background,
)
from idtrackerai.GUI_tools import Canvas, WrappedLabel
from idtrackerai.utils import Episode, IdtrackeraiError


def is_standard(stat: str) -> bool:
    return stat.lower() in ("median", "mean", "max", "min")


class BkgComputationThread(QThread):
    set_progress_value = Signal(int)
    set_progress_max = Signal(int)
    background_stat: str
    n_frames_for_background: int
    video_paths: Sequence[str | Path]
    episodes: list[Episode]

    def __init__(self):
        super().__init__()
        self.frame_stack = None
        self.bkg = None
        self.abort = False

    def setStat(self, stat: str):
        new_stat = stat.lower()
        self.background_stat = new_stat
        self.bkg = None
        if hasattr(self, "video_paths"):
            # when the App in inactive, Stat is set but there is no video_paths yet
            if not is_standard(self.background_stat):
                # we load the custom background in the main thread because it can raise IdtrackeraiError
                self.bkg = load_custom_background(
                    self.background_stat, self.video_paths[0]
                )
                self.finished.emit()
            else:
                self.start()

    def set_parameters(
        self, video_paths: Sequence[str | Path], episodes: list[Episode]
    ):
        self.video_paths = video_paths
        self.episodes = episodes

    def run(self):
        self.abort = False
        if self.bkg is not None:
            return
        self.set_progress_max.emit(self.n_frames_for_background)
        if self.frame_stack is None:
            self.frame_stack = generate_frame_stack(
                self.episodes,
                self.n_frames_for_background,
                self.set_progress_value,
                lambda: self.abort,
            )
        if self.abort:
            self.frame_stack = None
            self.abort = False
            return

        self.set_progress_value.emit(0)
        self.set_progress_max.emit(0)
        self.bkg = generate_background_from_frame_stack(
            self.frame_stack, self.background_stat
        )

        if self.abort:
            self.frame_stack = None
            self.bkg = None
            self.abort = False
            return

    def quit(self):
        self.abort = True


class ImageDisplay(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Background")
        self.canvas = Canvas()
        self.canvas.painting_time.connect(self.paint_image)

        self.setLayout(QHBoxLayout())
        self.setMinimumSize(200, 50)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self.canvas)

    def paint_image(self, painter: QPainter):
        painter.drawPixmap(0, 0, self.pixmap)

    def show(self, frame: np.ndarray):
        height, width = frame.shape
        self.pixmap = QPixmap.fromImage(
            QImage(frame.data, width, height, width, QImage.Format.Format_Grayscale8)
        )

        self.canvas.centerX = int(width / 2)
        self.canvas.centerY = int(height / 2)

        ratio = width / height

        QDialog_size = 600
        if width > height:
            window_width = QDialog_size
            window_height = int(QDialog_size / ratio)
        else:
            window_width = int(QDialog_size / ratio)
            window_height = QDialog_size
        self.setGeometry(0, 0, window_width, window_height)
        QTimer.singleShot(0, lambda: self.canvas.adjust_zoom_to(width, height))
        super().exec()


class BkgWidget(QWidget):
    new_bkg_data = Signal(object)  # np.ndarray | None

    def __init__(self):
        super().__init__()
        self.checkBox = QCheckBox("Background\nsubtraction")
        self.checkBox.stateChanged.connect(self.CheckBox_changed)

        self.bkg_stat = QComboBox()
        self.bkg_stat.addItems(("Custom", "Median", "Mean", "Max", "Min"))
        self.bkg_stat.setCurrentIndex(-1)
        self.bkg_stat.setEnabled(False)
        self.bkg_stat.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.bkg_stat.setSizePolicy(
            QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum
        )
        self.last_stat = "Median"

        self.uploaded_path = WrappedLabel(framed=True)
        self.uploaded_path.clicked.connect(lambda: self.bkg_stat_changed("Custom"))
        self.uploaded_path.setVisible(False)

        self.view_bkg = QToolButton()
        self.view_bkg.setText("View background")
        self.bkg_thread = BkgComputationThread()
        self.view_bkg.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.view_bkg.setEnabled(False)
        self.view_bkg.clicked.connect(self.view_bkg_clicked)

        self.image_display = ImageDisplay(self)
        layout = QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.checkBox)
        layout.addWidget(self.bkg_stat)
        layout.addWidget(self.uploaded_path)
        layout.addWidget(self.view_bkg)
        self.bkg_stat.currentTextChanged.connect(self.bkg_stat_changed)
        self.bkg_thread.set_progress_value.connect(self.set_progress_value)
        self.bkg_thread.set_progress_max.connect(self.set_progress_maximum)
        self.bkg_thread.started.connect(self.bkg_thread_started)
        self.bkg_thread.finished.connect(self.bkg_thread_finished)

    def bkg_stat_changed(self, text: str) -> None:
        if is_standard(text):
            self.bkg_thread.setStat(text)
            return

        assert text.lower() == "custom"

        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Select background image",
            self.uploaded_path.text(),
            "Image files (*.png *.jpg *.jpeg *.bmp *.tiff);;All files (*)",
        )
        if not file_path:
            self.set_bkg_stat(self.last_stat)
            return

        self.set_custom_path(file_path)

    def set_custom_path(self, file_path: str):
        try:
            self.bkg_thread.setStat(file_path)
        except IdtrackeraiError as err:
            QMessageBox.critical(None, "Error", str(err))
            self.set_bkg_stat(self.last_stat)
            return

        self.uploaded_path.setText(file_path)
        self.bkg_stat.blockSignals(True)
        self.bkg_stat.setCurrentText("Custom")
        self.bkg_stat.blockSignals(False)

    def set_bkg_stat(self, stat: str):
        if stat.lower() in ("median", "mean", "max", "min", "custom"):
            self.bkg_stat.setCurrentText(stat.capitalize())
        else:
            self.set_custom_path(stat)

    def set_new_video_paths(self, video_paths, episodes):
        self.video_paths = video_paths
        self.episodes = episodes
        self.bkg_thread.bkg = None
        self.bkg_thread.frame_stack = None
        if not self.checkBox.isChecked() or self.bkg_stat.currentText() == "Custom":
            return
        QMessageBox.information(
            self,
            "Background deactivated",
            "The subtracted background depends on the specified video paths. Check"
            " again the background subtraction if desired when finish editing the"
            " video paths.",
        )
        self.checkBox.setChecked(False)

    def view_bkg_clicked(self):
        if self.bkg_thread.bkg is not None:
            self.image_display.show(self.bkg_thread.bkg)

    def CheckBox_changed(self, checked):
        if checked:
            if not hasattr(self, "video_paths"):
                self.checkBox.setChecked(False)
                return
            self.bkg_thread.set_parameters(self.video_paths, self.episodes)
            self.bkg_thread.setStat(
                self.bkg_stat.currentText()
                if is_standard(self.bkg_stat.currentText())
                else self.uploaded_path.text()
            )
        else:
            self.view_bkg.setEnabled(False)
            self.new_bkg_data.emit(None)
        self.bkg_stat.setEnabled(checked)
        self.uploaded_path.setEnabled(checked)

    def bkg_thread_started(self):
        self.progress_bar = QProgressDialog("Computing background", "Cancel", 0, 100)
        self.progress_bar.setMinimumDuration(200)
        self.progress_bar.setModal(True)
        self.progress_bar.setAutoReset(False)
        self.progress_bar.canceled.connect(self.bkg_thread.quit)

    def set_progress_value(self, value: int):
        self.progress_bar.setValue(value)

    def set_progress_maximum(self, value: int):
        self.progress_bar.setMaximum(value)

    def bkg_thread_finished(self):
        if hasattr(self, "progress_bar"):
            self.progress_bar.reset()
        if self.bkg_thread.bkg is None:
            self.checkBox.setChecked(False)
        else:
            self.view_bkg.setEnabled(True)
            self.last_stat = (
                self.bkg_stat.currentText()
                if is_standard(self.bkg_stat.currentText())
                else self.uploaded_path.text()
            )
        self.uploaded_path.setVisible(not is_standard(self.last_stat))
        self.new_bkg_data.emit(self.bkg_thread.bkg)

    def getBkg(self):
        return self.bkg_thread.bkg if self.checkBox.isChecked() else None

    def get_bkg_stat(self) -> None | str:
        if not self.checkBox.isChecked():
            return None
        if self.bkg_stat.currentText() == "Custom":
            return self.uploaded_path.text()
        return self.bkg_stat.currentText().lower()



from collections.abc import Sequence

import numpy as np
from qtpy.QtCore import QPoint, Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QWidget

from idtrackerai.base.animals_detection import process_frame
from idtrackerai.GUI_tools import CanvasPainter


class FrameAnalyzer(QWidget):
    new_areas = Signal(int, list)
    new_parameters = Signal()

    def set_bkg(self, bkg_model):
        self.bkg_model = bkg_model
        self.use_bkg = bkg_model is not None
        self.need_to_redraw = True
        self.new_parameters.emit()

    def set_ROI_mask(self, ROI_mask: np.ndarray | None):
        self.ROI_mask = ROI_mask
        self.need_to_redraw = True
        self.new_parameters.emit()

    def set_intensity_ths(self, intensity_ths: Sequence[float]):
        self.intensity_ths = intensity_ths
        self.need_to_redraw = True
        self.new_parameters.emit()

    def set_area_ths(self, area_ths: Sequence[float]):
        self.area_ths = area_ths
        self.need_to_redraw = True
        self.new_parameters.emit()

    def __init__(self):
        super().__init__()

        self.use_bkg = False
        self.bkg_model = None
        self.ROI_mask = None
        self.intensity_ths = [0, 1]
        self.area_ths = [1, 1]
        self.blob_polygons: list[list[QPoint]] = []
        self.drawn_frame = -1

    def process_frame(self, frame: np.ndarray | None):
        if frame is None:  # error frame
            self.areas, contours = [], []
        else:
            self.areas, contours, gray_frame = process_frame(
                frame,
                bkg_model=self.bkg_model,
                ROI_mask=self.ROI_mask,
                intensity_ths=self.intensity_ths,
                area_ths=self.area_ths,
            )

        self.n_blobs = len(contours)
        for i, contour in enumerate(contours):
            polygon = [QPoint(*xy) for xy in contour]
            if i == len(self.blob_polygons):
                self.blob_polygons.append(polygon)
            else:
                self.blob_polygons[i] = polygon

    def paint_on_canvas(
        self, painter: CanvasPainter, frame_number: int, frame: np.ndarray | None
    ):
        if self.drawn_frame != frame_number or self.need_to_redraw:
            self.process_frame(frame)
            self.new_areas.emit(frame_number, self.areas)
            self.need_to_redraw = False
        painter.setBrush(QColor(60, 160, 255, 150))
        painter.setPen(QColor(0x286384))
        for i in range(self.n_blobs):
            painter.drawPolygon(self.blob_polygons[i])  # type: ignore
        self.drawn_frame = frame_number



from qtpy.QtCore import QPointF, Qt, Signal  # type: ignore[reportPrivateImportUsage]
from qtpy.QtGui import QColor, QColorConstants
from qtpy.QtWidgets import QInputDialog, QVBoxLayout, QWidget

from idtrackerai.GUI_tools import (
    AddBtn,
    CanvasMouseEvent,
    CanvasPainter,
    CustomList,
    LightPopUp,
    point_colors,
)
from idtrackerai.utils import LengthCalibration


class LengthCalibrator(QWidget):
    needToDraw = Signal()
    current_calibration: LengthCalibration | None = None

    def __init__(self) -> None:
        super().__init__()

        self.add = AddBtn()
        self.list = CustomList(max_n_row=0)

        layout = QVBoxLayout()
        layout.setSpacing(2)
        self.setLayout(layout)
        layout.addWidget(self.add)
        layout.addWidget(self.list)

        self.add.clicked.connect(self.add_clicked)
        self.calibrations: list[LengthCalibration] = []
        self.list.ListChanged.connect(self.needToDraw.emit)
        self.list.removedItem.connect(self.remove_item)
        self.color_count = -1
        self.popup = LightPopUp()
        self.info_dialog_already_displayed: bool = False

    def click_event(self, event: CanvasMouseEvent) -> None:
        if (
            not self.isVisible()
            or not self.isEnabled()
            or self.current_calibration is None
        ):
            return

        if event.button == Qt.MouseButton.LeftButton:
            self.current_calibration.add_point(event.int_xy_data)
            if self.current_calibration.has_two_points():
                self.end_calibration()

    def end_calibration(self) -> None:
        if self.current_calibration is None:
            return

        valid = False
        dialog_text = "Enter real distance between the two points:"
        while not valid:
            value_str, ok = QInputDialog.getText(self.add, "idtracker.ai", dialog_text)
            if not ok:
                self.current_calibration = None
                return
            try:
                value = float(value_str)
            except ValueError:
                valid = False
            else:
                valid = True
            if not valid:
                dialog_text = "Invalid characters encountered"

        self.current_calibration.distance = value
        self.calibrations.append(self.current_calibration)
        self.list.add_str(
            str(self.current_calibration), color=QColor(self.current_calibration.color)
        )
        self.current_calibration = None

    def add_clicked(self) -> None:
        if not self.info_dialog_already_displayed:
            self.popup.info(
                "Length Calibration",
                "Indicate two points by clicking in the video and then enter the"
                " numeric value of the distance between these two. The average factor"
                " of all calibrations will be included in the trajectory files.",
            )
            self.info_dialog_already_displayed = True

        self.color_count = (
            0 if self.color_count == len(point_colors) - 1 else self.color_count + 1
        )
        self.current_calibration = LengthCalibration(point_colors[self.color_count])
        self.needToDraw.emit()

    def remove_item(self, data: str) -> None:
        for calibration in self.calibrations:
            if str(calibration) == data:
                self.calibrations.remove(calibration)
                return

    def load(self, calibrations: list[LengthCalibration] | None) -> None:
        self.list.clear()
        if calibrations is None:
            return

        for calibration in calibrations:
            self.color_count = (
                0 if self.color_count == len(point_colors) - 1 else self.color_count + 1
            )
            calibration.color = point_colors[self.color_count]
            self.calibrations.append(calibration)
            self.list.add_str(str(calibration), color=QColor(calibration.color))
        self.current_calibration = None

    def get_calibrations(self) -> list[LengthCalibration]:
        return [c for c in self.calibrations if c.completed()]

    def paint_on_canvas(self, painter: CanvasPainter) -> None:
        painter.setPen(QColorConstants.Black)
        for calibration in self.calibrations:
            assert calibration.point_A is not None
            assert calibration.point_B is not None
            painter.setPen(QColor(calibration.color))
            painter.drawLine(*(calibration.point_A + calibration.point_B))  # type: ignore
            painter.setPen(QColorConstants.Black)
            painter.drawText(
                QPointF(
                    0.5 * (calibration.point_A[0] + calibration.point_B[0]),
                    0.5 * (calibration.point_A[1] + calibration.point_B[1]),
                ),
                f"{calibration.distance}",
            )
            painter.setBrush(QColor(calibration.color))
            painter.drawBigPoint(*calibration.point_A)
            painter.drawBigPoint(*calibration.point_B)

        calibration = self.current_calibration
        if calibration is None:
            return
        painter.setBrush(QColor(calibration.color))

        if calibration.point_A is not None:
            painter.drawBigPoint(*calibration.point_A)

        if calibration.point_B is not None:
            painter.drawBigPoint(*calibration.point_B)



"Identification of individual fragments given the predictions generate by the idCNN"

import json
import logging
from collections.abc import Sequence
from pathlib import Path

import numpy as np
import torch

from idtrackerai import Fragment, ListOfFragments
from idtrackerai.utils import track

from ..network import IdentifierBase, get_predictions


def compute_identification_statistics_for_non_accumulated_fragments(
    fragments: Sequence[Fragment],
    all_predictions: np.ndarray,
    all_probabilities: np.ndarray,
    number_of_animals: int,
):
    """Given the predictions associated to the images in each (individual)
    fragment in the list fragments if computes the statistics necessary for the
    identification of fragment.

    Parameters
    ----------
    fragments : list
        List of individual fragment objects
    assigner : <GetPrediction object>
        The assigner object has as main attributes the list of predictions
        associated to `images` and the the corresponding softmax vectors
    number_of_animals : int
        number of animals to be tracked
    """
    counter = 0
    for fragment in fragments:
        next_counter_value = counter + fragment.n_images
        predictions = all_predictions[counter:next_counter_value]
        probabilities = all_probabilities[counter:next_counter_value]
        fragment.set_identification_statistics(
            predictions, probabilities, number_of_animals
        )
        counter = next_counter_value
    assert counter == len(all_predictions)


def check_penultimate_model(
    identification_model: torch.nn.Module,
    model_path: Path,
    penultimate_model_path: Path,
):
    """Loads the penultimate accumulation step if the validation accuracy of the last
    step was lower then the penultimate. This discard possible corrupt final accumulation steps
    """
    if not penultimate_model_path.is_file():
        return

    last_model: dict = json.loads(model_path.with_suffix(".metadata.json").read_text())
    last_accuracy = last_model.get("test_acc", 0.0)
    last_ratio_accumulated = last_model.pop("ratio_accumulated", 0.0)
    penultimate_model: dict = json.loads(
        penultimate_model_path.with_suffix(".metadata.json").read_text()
    )
    penultimate_accuracy = penultimate_model.pop("test_acc", -1.0)
    penultimate_ratio_accumulated = penultimate_model.pop("ratio_accumulated", -1.0)
    logging.info(
        f"Last accuracy = {last_accuracy:.2%}, "
        f"Last ratio accumulated = {last_ratio_accumulated:.2%}\n"
        f"Penultimate accuracy = {penultimate_accuracy:.2%}, "
        f"Penultimate ratio accumulated = {penultimate_ratio_accumulated:.2%}"
    )

    if penultimate_ratio_accumulated < 0.9:
        logging.info(
            "The penultimate accumulation step had a ration of accumulated images lower"
            " than 90%. Thus, it will be ignored."
        )
        return

    if penultimate_accuracy > last_accuracy:
        logging.info(
            "The last accumulation step had a lower accuracy than the penultimate."
        )
        logging.info("Loading penultimate model, %s", penultimate_model_path)
        identification_model.load_state_dict(
            torch.load(penultimate_model_path, weights_only=True)
        )
    else:
        logging.info(
            "The last accumulation step had a higher accuracy than the penultimate."
        )


def assign_remaining_fragments(
    list_of_fragments: ListOfFragments, identification_model: IdentifierBase
):
    """This is the main function of this module: given a list_of_fragments it
    puts in place the routine to identify, if possible, each of the individual
    fragments. The starting point for the identification is given by the
    predictions produced by the ConvNetwork net passed as input. The organisation
    of the images in individual fragments is then used to assign more accurately.

    Parameters
    ----------
    list_of_fragments : <ListOfFragments object>
        collection of the individual fragments and associated methods
    session : <Session object>
        Object collecting all the parameters of the video and paths for saving and loading
    net : <ConvNetwork object>
        Convolutional neural network object created according to net.params

    See Also
    --------
    ListOfFragments.get_images_from_fragments_to_assign
    assign
    compute_identification_statistics_for_non_accumulated_fragments

    """
    logging.info("Assigning identities to all non-accumulated individual fragments")
    list_of_fragments.reset(roll_back_to="accumulation")
    fragments_to_identify = [
        frag
        for frag in list_of_fragments.individual_fragments
        if not frag.used_for_training
    ]

    logging.info(
        f"Number of unidentified individual fragments: {len(fragments_to_identify)}"
    )
    if not fragments_to_identify:
        list_of_fragments.compute_P2_vectors()
        return

    image_locations: list[tuple[int, int]] = []
    for fragment in fragments_to_identify:
        image_locations += fragment.image_locations

    logging.info(
        "Number of images to identify non-accumulated fragments: %d",
        len(image_locations),
    )

    predictions, probabilities = get_predictions(
        identification_model, image_locations, list_of_fragments.id_images_file_paths
    )

    logging.debug(
        f"{len(predictions)} generated predictions between "
        f"{len(set(predictions))} identities"
    )
    compute_identification_statistics_for_non_accumulated_fragments(
        fragments_to_identify, predictions, probabilities, list_of_fragments.n_animals
    )

    list_of_fragments.compute_P2_vectors()
    for fragment in track(
        list_of_fragments.get_fragments_to_identify(),
        "Assigning remaining fragments identities",
        sum(frag.identity is None for frag in list_of_fragments.individual_fragments),
    ):
        fragment.assign_identity(
            list_of_fragments.n_animals, list_of_fragments.id_to_exclusive_roi
        )

    list_of_fragments.compute_P2_vectors()



"Functions to manage logging and exception handling"

import logging
import logging.handlers
import multiprocessing
import os
import threading
from collections.abc import Callable
from contextlib import suppress
from datetime import datetime
from functools import wraps
from importlib import metadata
from pathlib import Path
from platform import platform, python_version
from shutil import copyfileobj
from tempfile import TemporaryFile
from traceback import extract_tb

from packaging.version import Version
from rich.console import Console, ConsoleRenderable
from rich.logging import RichHandler

from .py_utils import IdtrackeraiError, resolve_path
from .telemetry import (
    check_version,
    check_version_on_console_thread,
    report_usage_on_console_thread,
)

# The logging from other processes is sent to the listener thread of
# the main process by using the LOGGING_QUEUE.
# This queue is created in the main process and passed to the other processes.
if multiprocessing.current_process().name == "MainProcess":
    LOGGING_QUEUE = multiprocessing.Queue()
else:
    LOGGING_QUEUE = None

LOG_FILE_PATH = resolve_path("idtrackerai.log")

# This temporary file stores the log until the tracking is over
# and its contents are copied into the session folder log file
TMP_LOG_FILE = TemporaryFile("w+", suffix=".idtrackerai.log", encoding="utf_8")

ERROR_MSG = (
    "\n{}\nIf this error happens right after the installation,"
    " check our installation troubleshooting guide"
    " https://idtracker.ai/latest/install/installation_troubleshooting.html"
    "\n\nIf this error persists please let us know by following any of the following"
    " options:\n  - Posting on https://groups.google.com/g/idtrackerai_users\n  -"
    " Opening an issue at https://gitlab.com/polavieja_lab/idtrackerai\n  - Sending an"
    " email to info@idtracker.ai\nShare the log file ({}) when doing"
    " any of the options above"
)

LEVEL_FORMAT = {
    "WARNING": "red",
    "WARN": "red",
    "ERROR": "red bold",
    "FATAL": "red bold reverse",
    "CRITICAL": "red bold reverse",
    "NOTSET": "red",
}


class LevelRichHandler(RichHandler):
    """Modified Rich Logging Handler that prints the logging level only if its
    equal or grater than a warning. Intended to work with `show_level=False`"""

    def render_message(
        self, record: logging.LogRecord, message: str
    ) -> ConsoleRenderable:
        """Method override to manage log level only if its >= warning

        Args:
            record (LogRecord): logging Record.
            message (str): String containing log message.

        Returns:
            ConsoleRenderable: Renderable to display log message.
        """
        if record.levelno >= logging.WARNING:
            record.markup = True
            format = LEVEL_FORMAT.get(record.levelname, "red")
            message = f"[{format}]{record.levelname}[/{format}] {message}"
        return super().render_message(record, message)


def _log_listener_thread(queue: multiprocessing.Queue) -> None:
    logger = logging.getLogger()
    while True:
        record = queue.get()
        if record is None:
            # We send this as a sentinel to tell the listener to quit.
            break
        logger.handle(record)


def setup_logging_queue(queue: multiprocessing.Queue) -> None:
    """
    Intended to be run in a child process. Sets up a logging queue
    to send log messages to the listener thread in the main process.
    """
    if queue is None:
        # This is the main process, nothing to do
        return
    logging_queue_handler = logging.handlers.QueueHandler(queue)
    root = logging.getLogger()
    root.setLevel(logging.NOTSET)
    if any(isinstance(handler, LevelRichHandler) for handler in root.handlers):
        # When multiprocessing with 'fork' method, the child process inherits the parent's handlers
        # We do not want to add the handler twice in this case
        return
    root.handlers.clear()
    root.addHandler(logging_queue_handler)


def init_logger(level: int = logging.DEBUG, write_to_disk: bool = False) -> None:
    """Initializes the logger with a custom re-implementation of Rich logging handler"""
    global LOG_FILE_PATH
    logger_width_when_no_terminal = 126
    try:
        os.get_terminal_size()
    except OSError:
        # stdout is sent to file. We define logger width to a constant
        size = logger_width_when_no_terminal
    else:
        # stdout is sent to terminal
        # We define logger width to adapt to the terminal width
        size = None

    # The first handler is the terminal, the second one the .log file,

    handlers = [
        LevelRichHandler(
            console=Console(width=size), rich_tracebacks=True, show_level=False
        )
    ]

    if write_to_disk:
        for counter in range(1, 21):  # try 30 different log file names...
            try:
                LOG_FILE_PATH.unlink(True)  # avoid conflicts and merged files
            except PermissionError as exc:
                # In Windows this arises when the log file is open by another session
                logging.warning(
                    f"Could not set up the logging file {LOG_FILE_PATH} because "
                    f"of the following PermissionError: {exc}"
                )
                LOG_FILE_PATH = LOG_FILE_PATH.with_stem(f"idtrackerai_{counter}")
            else:
                break
        else:
            logging.error("Could not set up the logging file")

        handlers.extend(
            (
                LevelRichHandler(
                    console=Console(
                        file=LOG_FILE_PATH.open("w", encoding="utf_8"),  # noqa SIM115
                        width=logger_width_when_no_terminal,
                    ),
                    show_level=False,
                ),
                LevelRichHandler(
                    console=Console(
                        file=TMP_LOG_FILE, width=logger_width_when_no_terminal
                    ),
                    show_level=False,
                ),
            )
        )

    logging.basicConfig(
        level=level,
        format="%(message)s",
        datefmt="%H:%M:%S",
        force=True,
        handlers=handlers,
    )

    os.environ["OPENCV_FFMPEG_LOGLEVEL"] = "-8"  # avoid huge logs with corrupted videos
    logging.getLogger("urllib3").setLevel(logging.INFO)
    logging.captureWarnings(True)
    logging.info(
        f"[bold]Welcome to idtracker.ai[/] {metadata.version('idtrackerai')}",
        extra={"markup": True},
    )
    logging.debug(
        f"Date: {str(datetime.now()).split('.')[0]}\n"
        f"Running Python '{python_version()}' in '{platform(True)}'"
    )
    if write_to_disk:
        logging.info("Writing log in %s", LOG_FILE_PATH)
    else:
        logging.info("Not writing log in any file")
    logging.info("Using NumPy %s", metadata.version("numpy"))

    try:
        logging.info(
            "Using headless OpenCV %s", metadata.version("opencv-python-headless")
        )
    except metadata.PackageNotFoundError:
        logging.info("Headless OpenCV not found")

    try:
        logging.info("Using regular OpenCV %s", metadata.version("opencv-python"))
    except metadata.PackageNotFoundError:
        logging.info("Regular OpenCV not found")

    logging.info(
        'Using "%s" multiprocessing start method', multiprocessing.get_start_method()
    )

    if LOGGING_QUEUE:
        # We start the listener thread which will receive log messages from other processes
        threading.Thread(target=_log_listener_thread, args=(LOGGING_QUEUE,)).start()


def wrap_entrypoint(main_function: Callable):
    """Function decorator for CLI entry points.
    It initializes the logger, checks the software version and manages possible Exceptions
    """

    @wraps(main_function)
    def ret_fun(*args, **kwargs):
        init_logger(write_to_disk=True)
        check_version_on_console_thread()
        report_usage_on_console_thread()
        try:
            return_value = main_function(*args, **kwargs)
        except (Exception, KeyboardInterrupt) as exc:
            manage_exception(exc)
            if hasattr(exc, "log_path"):
                TMP_LOG_FILE.flush()
                TMP_LOG_FILE.seek(0)
                with open(exc.log_path, "w", encoding="utf-8") as file:  # type: ignore
                    copyfileobj(TMP_LOG_FILE, file)
                logging.info(f"Log file copied to {exc.log_path}")  # type: ignore
            return_value = False
        finally:
            if LOGGING_QUEUE:
                # We send this as a sentinel to tell the listener thread to quit.
                LOGGING_QUEUE.put(None)
        return return_value

    return ret_fun


def manage_exception(exc: BaseException) -> None:
    """Prints useful log messages depending on the type of Exception"""
    kind, message = check_version()
    ERROR_MSG_ = ERROR_MSG.format(
        message + "\n" if kind.update_available else "", LOG_FILE_PATH
    )
    match exc:
        case IdtrackeraiError():
            tb = extract_tb(exc.__traceback__)[-1]
            logging.critical(
                "%s [bright_black](from %s:%d)[/]",
                exc,
                Path(*Path(tb.filename).parts[-2:]),
                tb.lineno,
                extra={"markup": True},
            )

        case KeyboardInterrupt():
            logging.critical("KeyboardInterrupt", exc_info=False)

        case ModuleNotFoundError():
            if "torch" in str(exc):
                logging.critical(
                    "Module PyTorch is not installed, follow their guideline to install"
                    " it (https://pytorch.org/get-started/locally/). Original"
                    ' exception: "%s"',
                    exc,
                )
                return
            logging.critical("%s: %s", type(exc).__name__, exc, exc_info=exc)
            logging.info(ERROR_MSG_)

        case RuntimeError():
            with suppress(metadata.PackageNotFoundError):
                if (Version(metadata.version("torch")) < Version("2.3")) and (
                    Version(metadata.version("numpy")) >= Version("2.0")
                ):
                    logging.error(str(exc))
                    logging.critical(
                        "This error may be caused by your PyTorch installation (version %s) being incompatible with NumPy 2.0 or higher, please update PyTorch by running the installation command in https://pytorch.org/get-started/locally/#start-locally",
                        metadata.version("torch"),
                    )
                    return

            logging.critical("%s: %s", type(exc).__name__, exc, exc_info=exc)
            logging.info(ERROR_MSG_)

        case Exception():
            logging.critical("%s: %s", type(exc).__name__, exc, exc_info=exc)
            logging.info(ERROR_MSG_)



from .GUI_main_base import GUIBase
from .widgets_utils.canvas import Canvas, CanvasMouseEvent, CanvasPainter
from .widgets_utils.custom_list import CustomList
from .widgets_utils.id_labels import IdLabels
from .widgets_utils.other_utils import (
    AddBtn,
    LightPopUp,
    QHLine,
    RemoveBtn,
    TransparentDisabledOverlay,
    WrappedLabel,
    build_ROI_patches_from_list,
    get_icon,
    get_path_from_points,
    key_event_modifier,
    open_session,
)
from .widgets_utils.sliders import InvertibleSlider, LabelRangeSlider
from .widgets_utils.video_paths_holder import VideoPathHolder
from .widgets_utils.video_player import VideoPlayer

point_colors: list[int] = [
    0x9467BD,
    0x2CA02C,
    0xBCBD22,
    0xFF7F0E,
    0x8C564B,
    0xE377C2,
    0x7F7F7F,
    0x17BECF,
]


__all__ = [
    "point_colors",
    "IdLabels",
    "open_session",
    "LabelRangeSlider",
    "CustomList",
    "WrappedLabel",
    "Canvas",
    "CanvasPainter",
    "GUIBase",
    "VideoPlayer",
    "VideoPathHolder",
    "key_event_modifier",
    "build_ROI_patches_from_list",
    "QHLine",
    "CanvasMouseEvent",
    "get_path_from_points",
    "LightPopUp",
    "InvertibleSlider",
    "TransparentDisabledOverlay",
    "AddBtn",
    "RemoveBtn",
    "get_icon",
]
